"""Module implements Baseclasses."""

from __future__ import division, print_function, absolute_import

from abc import ABCMeta, abstractmethod
from six import add_metaclass

from SafeRLBench import AlgoMonitor, EnvMonitor

__all__ = ('EnvironmentBase', 'Space')


@add_metaclass(ABCMeta)
class EnvironmentBase(EnvMonitor):
    """Environment Base Class.

    This base class defines and implements an interface to any environment
    implementation part of the environment module. Subclasses inheriting
    from EnvironmentBase need to make sure they meet the requirements below.

    Any subclass must implement:
        * _update(action)
        * _reset()

    Any subclass might override:
        * _rollout(policy)

    Make sure the `state_space`, `action_space` and `horizon` attributes will
    be set in any subclass, as the default implementation and / or the monitor
    may access them to retrieve information.

    Attributes
    ----------
    state_space :
        State space of the environment.
    action_space :
        Action space of the environment.
    horizon :
        Maximum number of iterations until rollout will stop.
    monitor : EnvData instance
        Contains the monitoring data. The monitor will be automatically
        initialized during creation.

    Methods
    -------
    rollout(policy)
        Perform a rollout according to the actions selected by policy.
    update(action)
        Update the environment state according to the action.
    reset()
        Reset the environment to the initial state.

    Notes
    -----
    When overwriting _rollout(policy) use the provided interface functions
    and do not directly call the private implementation.
    """

    def __init__(self, state_space, action_space, horizon=0):
        """Initialize EnvironmentBase.

        Parameters
        ----------
        state_space :
            State space of the environment.
        action_space :
            Action space of the environment.
        horizon :
            Maximum number of iterations until rollout will stop.
        """
        super(EnvironmentBase, self).__init__()
        self.state_space = state_space
        self.action_space = action_space
        self.horizon = horizon

    # Implement in subclasses:
    # See update(self, action) for more information
    @abstractmethod
    def _update(self, action):
        raise NotImplementedError

    # See reset(self) for more information
    @abstractmethod
    def _reset(self):
        raise NotImplementedError

    # Override in subclasses if necessary
    def _rollout(self, policy):
        self.reset()
        trace = []
        for n in range(self.horizon):
            action = policy(self.state)
            trace.append(self.update(action))
        return trace

    def update(self, action):
        """Update the environment state according to the action.

        Wraps the subclass implementation _update(action) providing
        monitoring capabilities.

        Parameters
        ----------
        action: array-like
            Element of action_space

        Returns
        -------
        tuple : 3-tuple
            action : array-like
                element of action space as it has been applied in update
            state : array-like
                element of state_space which is the resulting state after
                applying action
            reward : float
                reward for resulting state
        """
        with self.monitor_update():
            t = self._update(action)
        return t

    def reset(self):
        """Reset the environment to initial state.

        Reset wraps the subclass implementation _reset() providing monitoring
        capabilities.
        """
        with self.monitor_reset():
            self._reset()

    def rollout(self, policy):
        """Perform a rollout according to the actions selected by policy.

        Wraps the implementation _rollout(policy) providing monitoring
        capabilities.

        Parameters
        ----------
        Policy : callable
            Maps element of state_space to element of action_space

        Returns
        -------
        trace : list of 3-tuple
            List of (action, state, reward)-tuple as returned by update().
        """
        with self.monitor_rollout():
            trace = self._rollout(policy)
        return trace

    def __repr__(self):
        """Return class name."""
        return self.__class__.__name__


@add_metaclass(ABCMeta)
class Space(object):
    """Baseclass for Spaceobject.

    All methods have to be implemented in any subclass.

    Methods
    -------
    contains(x)
        Check if x is an element of space.
    element
        Return arbitray element in space.
    """

    @abstractmethod
    def contains(self, x):
        """Check if x is an element of space."""
        pass

    @abstractmethod
    def sample(self):
        """Return an arbitrary element in space for unit testing."""
        pass

    @property
    @abstractmethod
    def dimension(self):
        """Return the dimension of the space."""
        pass


@add_metaclass(ABCMeta)
class AlgorithmBase(AlgoMonitor):
    """Baseclass for any algorithm.

    This baseclass defines a uniform interface for any algorithm part of
    the algorithm module SafeRLBench.algo. It features monitoring capabilities
    for tracking and evaluating the execution of the algorithm.

    Inheriting from `AlgorithmBase` is suspect to some constraints, i.e. any
    algorithm needs to be implemented using the following functions.

    Any subclass must overwrite:
        * _initialize(policy)
        * _step(policy)
        * _is_finished()

    Any subclass may overwrite:
        * _optimize(policy)

    In case one does overwrite _optimize, the functions _initialize(),
    _step(parameter), _is_finished() may just pass unless they are used.
    This may however change the information tracked by the monitor.

    Attributes
    ----------
    environment :
        Environment we want to optimize on
    policy :
        Policy to be optimized
    max_it : int
        Maximum number of iterations
    monitor : AlgoData instance
        Contains monitoring data. The monitor will automatically initialize
        on creation of an object.

    Methods
    -------
    optimize()
        Optimize a policy with respective algorithm.
    initialize()
        Initialize policy parameter.
    step()
        Update policy parameters.
    is_finished()
        Return true when algorithm is finished.

    Notes
    -----
    Specification of the private functions.

    _initialize(self):
        Initialize the algorithm.
    _step():
        Compute one step of the algorithm.
    _is_finished():
        Return True when algorithm is supposed to finish.
    """

    def __init__(self, environment, policy, max_it):
        super(AlgorithmBase, self).__init__()

        self.environment = environment
        self.policy = policy
        self.max_it = max_it

        self.grad = None

    # Have to be overwritten.
    @abstractmethod
    def _initialize(self):
        pass

    @abstractmethod
    def _step(self):
        pass

    @abstractmethod
    def _is_finished(self):
        pass

    # May be overwritten
    def _optimize(self):
        self.initialize()

        for n in range(self.max_it):
            self.step()
            if self.is_finished():
                break

    def optimize(self):
        """Optimize policy parameter.

        Wraps subclass implementation in _optimize(policy).

        Parameters
        ----------
        policy: PolicyBase subclass
        """
        with self.monitor_optimize():
            self._optimize()

    def initialize(self):
        """Initialize policy parameter.

        Wraps subclass implementation in _initialize(policy)

        Parameters
        ----------
        policy: PolicyBase subclass
        """
        with self.monitor_initialize():
            self._initialize()

    def step(self):
        """Update policy parameter.

        Wraps subclass implementation in _step(policy).

        Parameters
        ----------
        policy: PolicyBase subclass
        """
        with self.monitor_step():
            self._step()

    def is_finished(self):
        """Return True when algorithm is supposed to finish.

        Wraps subclass implementation in _is_finished().
        """
        stop = self._is_finished()
        return stop

    def reset(self):
        """Reset the monitor."""
        self._alg_reset()

    def __repr__(self):
        if hasattr(self, '_info'):
            return self._info()
        return self.__class__.__name__


@add_metaclass(ABCMeta)
class Policy(object):
    """Minimal policy interface."""

    def __call__(self, state):
        return self.map(state)

    @abstractmethod
    def map(self, state):
        """Map element of state space to action space."""
        pass

    @property
    @abstractmethod
    def parameters(self):
        """Current parameters."""
        pass

    @parameters.setter
    @abstractmethod
    def parameters(self, par):
        pass

    @property
    @abstractmethod
    def parameter_space(self):
        """Return parameter space."""


@add_metaclass(ABCMeta)
class ProbPolicy(Policy):
    """Probabilistic policy interface."""

    @abstractmethod
    def grad_log_prob(self, state, action):
        """Return the :math:log(grad p(action | state)):math:."""
        pass



"""Algorithm Tests."""

from SafeRLBench.algo import PolicyGradient, A3C
from SafeRLBench.envs import LinearCar
from .policygradient import CentralFDEstimator, estimators

from SafeRLBench.policy import NeuralNetwork

from unittest2 import TestCase
from mock import MagicMock, Mock


class TestPolicyGradient(TestCase):
    """PolicyGradientTestClass."""

    def test_pg_init(self):
        """Test: POLICYGRADIENT: initialization."""
        env_mock = MagicMock()
        pol_mock = Mock()

        for key, item in estimators.items():
            pg = PolicyGradient(env_mock, pol_mock, estimator=key)
            self.assertIsInstance(pg.estimator, item)

        pg = PolicyGradient(env_mock, pol_mock, estimator=CentralFDEstimator)
        self.assertIsInstance(pg.estimator, CentralFDEstimator)

        self.assertRaises(ImportError, PolicyGradient,
                          env_mock, pol_mock, CentralFDEstimator(env_mock))


class TestA3C(TestCase):
    """A3C Test Class."""

    def test_a3c_init(self):
        """Test: A3C: initialization."""
        a3c = A3C(LinearCar(), NeuralNetwork([2, 6, 1]))

        fields = ['environment', 'policy', 'max_it', 'num_workers', 'rate',
                  'done', 'policy', 'p_net', 'v_net', 'workers', 'threads',
                  'global_counter', 'sess']

        for field in fields:
            assert hasattr(a3c, field)



"""R^d with any shape."""
import numpy as np
from SafeRLBench import Space


class RdSpace(Space):
    """R^d Vectorspace."""

    def __init__(self, shape):
        """Initialize with shape."""
        self.shape = shape
        self._dim = None

    def contains(self, x):
        """Check if element is contained."""
        return isinstance(x, np.ndarray) and x.shape == self.shape

    def sample(self):
        """Return arbitrary element."""
        return np.ones(self.shape)

    @property
    def dimension(self):
        """Return dimension of the space."""
        if self._dim is None:
            d = 1
            for i in range(len(self.shape)):
                d *= self.shape[i]
            self._dim = d
        return self._dim

    def __repr__(self):
        return 'RdSpace(shape=%s)' % str(self.shape)



"""Benchmarking facilities."""

from SafeRLBench import EnvironmentBase, AlgorithmBase
from SafeRLBench import config

from concurrent.futures import ProcessPoolExecutor

from itertools import product

import logging
import pprint
try:
    from string import maketrans
except ImportError:
    def maketrans(a, b):
        return str.maketrans(a, b)

logger = logging.getLogger(__name__)

__all__ = ('Bench', 'BenchConfig')


def _dispatch_wrap(run):
    return Bench._dispatch(run)


class Bench(object):
    """Benchmarking class to benchmark algorithms on various environments.

    This class needs a ``BenchConfig`` instance that provides it with the
    necessary configurations. Refer to the ``BenchConfig`` documentations for
    instruction of how to set up a configurations.

    Attributes
    ----------
    config :
        Instance of ``BenchConfig``
    measures :
        List of measures

    Methods
    -------
    make_bench(algs, envs, measures=None)
        Static method that generates a configuration and returns an instance
        of Bench using this configuration.
    __call__()
        Benchmark then evaluate.
    benchmark()
        Initialize and run benchmark as configured.
    eval()
        Evaluate measures on test runs.

    Examples
    --------
    We can configure this class either in two steps, using the ``BenchConfig``
    class or in a single step using the ``make_bench`` method.

    Either way we start by defining our test runs. First define the environment
    configuration, e.g. LinearCar:

    >>> from SafeRLBench.envs import LinearCar
    >>> envs = [[(LinearCar, {'horizon': 100})]]

    Now we want to define a couple of different configurations for the
    PolicyGradient algorithm using list comprehension.

    >>> from SafeRLBench.algo import PolicyGradient
    >>> from SafeRLBench.policy import LinearPolicy
    >>> algs = [[
    ...     (PolicyGradient, [{
    ...         'policy': LinearPolicy(2, 1, par=[1, 1, 1]),
    ...         'estimator': 'central_fd',
    ...         'var': var
    ...     } for var in [1, 1.5, 2, 2.5]])
    ... ]]

    Now the first thing we could do is to manually generate a configuration and
    then use it to initialize ``Bench``.

    >>> from SafeRLBench import Bench, BenchConfig
    >>> config = BenchConfig(algs, envs)
    >>> bench = Bench(config)

    or we can do it in a single step using the static constructor in the
    ``Bench`` class.

    >>> # single step version is equivalent
    >>> bench = Bench.make_bench(algs, envs)

    To run the benchmarking and evaluation

    >>> bench()
    """

    def __init__(self, config=None, measures=None):
        """Initialize Bench instance.

        Parameters
        ----------
        configs : BenchConfig instance
            BenchConfig information supplying configurations. In case ``None``
            is passed an empty configuration will be instantiated. It can be
            populated with configurations using the ``add_tests`` methods of
            the configuration object. Default is ``None``.
        measures :
            List of measures from the measure module.
        """
        if not isinstance(config, BenchConfig):
            self.config = BenchConfig()
        else:
            self.config = config

        if measures is None:
            self.measures = []
        elif isinstance(measures, list):
            self.measures = measures
        else:
            self.measures = [measures]

        self.runs = []

    @staticmethod
    def make_bench(algs, envs, measures=None):
        """Construct a Bench directly.

        This will create the configuration instance and use it to create the
        bench. For detailed instructions on how the arguments work refer to
        the **Getting Started** section in the documentation or the
        ``BenchConfig`` documentation.

        Parameters
        ----------
        algs :
            List of list of tuples where the first element is an algorithm and
            the second a list of configurations. Any inner list may also be a
            single element instead of a list.
        envs :
            List of listof tuples where the first element is an environment and
            the second a configuration.
        measures :
            List of measures from the measure module.

        Returns
        -------
        bench : ``Bench`` object
            Confifigured ``Bench`` object.
        """
        config = BenchConfig(algs, envs)
        return Bench(config, measures)

    def __call__(self):
        """Initialize and run benchmark as configured."""
        self.benchmark()
        self.eval()

    def benchmark(self):
        """Initialize and run benchmark as configured."""
        logger.debug('Starting benchmarking.')

        self._set_up()

        if config.n_jobs > 1:
            self._benchmark_par()
        else:
            self._benchmark()

    def eval(self):
        """Evaluate measures on test runs."""
        for run in self.runs:
            if not run.completed:
                logger.warning("Evaluating before run completed.")

        for measure in self.measures:
            measure(self.runs)

    def _benchmark(self):
        for run in self.runs:
            self._dispatch(run)

    def _benchmark_par(self):
        n_jobs = config.n_jobs
        with ProcessPoolExecutor(max_workers=n_jobs) as ex:
            fs = [ex.submit(_dispatch_wrap, run) for run in self.runs]
            self.runs = [f.result() for f in fs]

    def _set_up(self):
        self.runs = []
        for alg, env, alg_conf, env_conf in self.config:
            env_obj = env(**env_conf)
            alg_obj = alg(env_obj, **alg_conf)

            self.runs.append(BenchRun(alg_obj, env_obj, alg_conf, env_conf))

    @staticmethod
    def _dispatch(run):
        logger.debug('DISPATCH RUN:\n\n%s\n', str(run))

        run.alg.optimize()
        run.completed = True

        return run


class BenchConfig(object):
    """Benchmark configuration class.

    This class is supposed to provide a convenient interface to setup
    configurations for benchmarking runs.

    Attributes
    ----------
    algs :
        List of list of tuples where the first element is an algorithm and
        the second a list of configurations. Any inner list may also be a
        single element instead of a list.

    envs :
        List of list of tuples where the first element is an environment and
        the second a configuration. Any inner list may also be a single element
        instead of a list.

    Notes
    -----
    When defining configurations for benchmarking the algorithm's configuration
    may depend on the respective environments configuration. Thus using a
    simple cartesian product of configurations may not be reasonable in many
    cases.

    To provide an easy to use interface, we thus assume two mayor cases:
    * One environment configuration for multiple algorithm configurations.
    * One algorithm configuration for multiple environment configurations.

    The outer lists of the environment and algorithm configurations need to
    have equal size, because they represent reasonable configuration pairs,
    i.e. they will be zipped.
    From the pairs of inner lists we then generate a grid (cartesian product)
    of the configuration tuples, representing the two cases described above.

    See the examples sections for examples regarding the two cases.

    Examples
    --------
    Let's say we want to benchmark multiple algorithms on one reference
    environment.

    First define the environment configuration, e.g. LinearCar:

    >>> from SafeRLBench.envs import LinearCar
    >>> envs = [[(LinearCar, {'horizon': 100})]]

    Now we want to define a couple of different configurations for the
    PolicyGradient algorithm using list comprehension.

    >>> from SafeRLBench.algo import PolicyGradient
    >>> from SafeRLBench.policy import LinearPolicy
    >>> algs = [[
    ...     (PolicyGradient, [{
    ...         'policy': LinearPolicy(2, 1, par=[1, 1, 1]),
    ...         'estimator': 'central_fd',
    ...         'var': var
    ...     } for var in [1, 1.5, 2, 2.5]])
    ... ]]

    Last we can retrieve an instance and use it to initialize the ``Bench``.

    >>> from SafeRLBench import Bench, BenchConfig
    >>> config = BenchConfig(algs, envs)
    >>> bench = Bench(config)
    """

    def __init__(self, algs=None, envs=None):
        """Initialize BenchConfig instance.

        This initializer may be used if you want to test multiple algorithm
        configurations on a list of environments.

        Parameters
        ----------
        algs :
            List of list of tuples where the first element is an algorithm and
            the second a list of configurations. Any inner list may also be a
            single element instead of a list.
        envs :
            List of listof tuples where the first element is an environment and
            the second a configuration.
        """
        if algs is None or envs is None:
            self.algs = []
            self.envs = []
            return

        if (len(algs) != len(envs)):
            raise ValueError('Configuration lists dont have same length')

        for i, obj in enumerate(algs):
            algs[i] = self._listify(obj)

        for i, obj in enumerate(envs):
            envs[i] = self._listify(obj)

        self.algs = algs
        self.envs = envs

    def add_tests(self, algs, envs):
        """Add one set of environment and algorithm configurations.

        Parameters
        ----------
        algs :
            List of tuples where the first element is an algorithm and the
            second a list of configurations. May also be single elements.
        envs :
            List of tuple where the first element is an environment and the
            second a list of configurations. May also be single elements.
        """
        algs = self._listify(algs)
        envs = self._listify(envs)

        self.algs.append(algs)
        self.envs.append(envs)

    def _listify(self, obj):

        if not isinstance(obj, list):
            obj = [obj]

        for i, tup in enumerate(obj):

            if not isinstance(tup, tuple):
                raise ValueError('Invalid input structure.')

            try:
                if not issubclass(tup[0], (AlgorithmBase, EnvironmentBase)):
                    raise TypeError
            except TypeError:
                raise ValueError('Invalid Algorithm or Environment')

            if not isinstance(tup[1], list):
                obj[i] = (tup[0], [tup[1]])

        return obj

    def __iter__(self):
        for algs, envs in zip(self.algs, self.envs):
            for (alg, alg_confs), (env, env_confs) in product(algs, envs):
                for alg_conf, env_conf in product(alg_confs, env_confs):
                    yield alg, env, alg_conf, env_conf


class BenchRun(object):
    """Wrapper containing instances and configuration for a run.

    Attributes
    ----------
    alg :
        Algorithm instance
    env :
        Environment instance
    alg_conf : Dictionary
        Algorithm configuration
    env_conf : Dictionary
        Environment configuration

    Methods
    -------
    get_alg_monitor()
        Retrieve the data container from the algorithm instance.
    get_env_monitor()
        Retrieve the data container from the environment instance.
    """

    def __init__(self, alg, env, alg_conf, env_conf):
        """Initialize BenchRun instance.

        Parameters
        ----------
        alg :
            Algorithm instance
        env :
            Environment instance
        alg_conf : Dictionary
            Algorithm configuration
        env_conf : Dictionary
            Environment configuration
        """
        self.alg = alg
        self.env = env

        self.alg_conf = alg_conf
        self.env_conf = env_conf

        self.completed = False

    def get_alg_monitor(self):
        """Retrieve AlgMonitor for algorithm."""
        return self.alg.monitor

    def get_env_monitor(self):
        """Retrieve EnvMonitor for environment."""
        return self.env.monitor

    def __repr__(self):
        out = []
        out += ['Algorithm: ', [self.alg.__class__.__name__, self.alg_conf]]
        out += ['Environment: ', [self.env.__class__.__name__, self.env_conf]]

        return pprint.pformat(out, indent=2).translate(maketrans(',\'[]',
                                                                 '    '))



"""Monitoring implementations."""

import logging
import time

from SafeRLBench import config

from contextlib import contextmanager

logger = logging.getLogger(__name__)

__all__ = ('EnvMonitor', 'AlgoMonitor')


class EnvMonitor(object):
    """
    Environment Monitor, providing tracking for environments.

    Attributes
    ----------
    monitor :
        This is the container where monitoring data will be stored.

    Methods
    -------
    monitor_update()
        Context manager for monitoring environment updates. It should be used
        when invoking the private ``_update`` implementation from the interface
        method.
    monitor_rollout()
        Context manager for monitoring environment rollout. It should be used
        when invoking the private ``_rollout`` implementation from the
        interface method.
    monitor_reset()
        Context manager for monitoring environment resets. It should be used
        when invoking the private ``_reset`` implementation from the interface
        method.
    """

    def __new__(cls, *args, **kwargs):
        """Hook into subclasses."""
        obj = object.__new__(cls)
        obj.monitor = EnvData()
        return obj

    @contextmanager
    def monitor_update(self):
        """Context monitoring update."""
        self._before_update()
        yield self
        self._after_update()

    @contextmanager
    def monitor_rollout(self):
        """Context monitoring rollout."""
        self._before_rollout()
        yield self
        self._after_rollout()

    @contextmanager
    def monitor_reset(self):
        """Context monitoring reset."""
        self._before_reset()
        yield self
        self._after_reset()

    def _before_update(self):
        """Monitor environment before update.

        Parameters
        ----------
        env :
            Environment instance to be monitored.
        """
        pass

    def _after_update(self):
        """Monitor environment after update.

        Parameters
        ----------
        env :
            Environment instance to be monitored.
        """
        pass

    def _before_rollout(self):
        """Monitor environment before rollout.

        Parameters
        ----------
        env :
            Environment instance to be monitored.
        """
        pass

    def _after_rollout(self):
        """
        Monitor environment after rollout.

        Parameters
        ----------
        env :
            Environment instance to be monitored.
        """
        self.monitor.rollout_cnt += 1

    def _before_reset(self):
        """Monitor environment before reset.

        Parameters
        ----------
        env :
            Environment instance to be monitored.
        """
        pass

    def _after_reset(self):
        """Monitor environment after reset.

        Parameters
        ----------
        env :
            Environment instance to be monitored.
        """
        pass


class AlgoMonitor(object):
    """Algorithm monitor tracks algorithms' activity.

    This class is inherited by the `AlgorithmBase` class and will provide it
    with tracking capabilities.

    Attributes
    ----------
    monitor :
        This is the container where monitoring data will be stored.
    grad :
        The Alogrithm can set this field, to provide information about the
        current gradient to the monitor.
    has_policy :
        In case the algorithm does not depend on a policy and does not need
        any parameters, this can me set to False, to prevent issues with
        tracking data that does not exist.

    Methods
    -------
    monitor_optimize()
        Context manager for monitoring algorithm optimizations. It should be
        used when invoking the private ``_optimize`` implementation from the
        interface method.
    monitor_initialize()
        Context manager for monitoring algorithm initializations. It should be
        used when invoking the private ``_initialize`` implementation from the
        interface method.
    monitor_step()
        Context manager for monitoring algorithm step. It should be used when
        invoking the private ``_step`` implementation from the interface
        method.
    """

    def __new__(cls, *args, **kwargs):
        """Hook into subclasses."""
        obj = object.__new__(cls)
        obj.monitor = AlgoData()
        obj.grad = None
        obj.has_policy = True
        return obj

    @contextmanager
    def monitor_optimize(self):
        """Context monitoring optimization."""
        self._before_optimize()
        yield self
        self._after_optimize()

    @contextmanager
    def monitor_initialize(self):
        """Context monitoring initialize."""
        yield self
        if self.has_policy:
            self.monitor.parameters.append(self.policy.parameters)

    @contextmanager
    def monitor_step(self):
        """Context monitoring stepping."""
        self._before_step()
        yield self
        self._after_step()

    def _before_optimize(self):
        """Setup montitor for optimization run.

        Parameters
        ----------
        alg :
            the algorithm instance to be monitored
        """
        if config.monitor_verbosity > 0:
            logger.info('Starting optimization of %s...', str(self))

        # reset monitor object in case of rerun
        self.monitor.reset()

        # init monitor dict for algorithm
        self.monitor.t = time.time()

        # init optimization time control
        self.monitor.optimize_start = time.time()

    def _after_optimize(self):
        """Catch data after optimization run."""
        # retrieve time of optimization
        optimize_end = time.time()
        optimize_time = optimize_end - self.monitor.optimize_start

        if self.monitor.optimize_start == 0:
            logger.warning('Time measure for optimize corrupted')

        self.monitor.optimize_start = 0

        self.monitor.optimize_time = optimize_time

        # if the gradient attribute has been set
        if self.grad is not None:
            logger.debug('Finished optimization after %d steps with grad %s.',
                         self.monitor.step_cnt, str(self.grad))
        else:
            logger.debug('Finished optimization after %d steps.',
                         self.monitor.step_cnt)

        if self.has_policy:
            # independently compute traces after optimization is finished
            if config.monitor_verbosity > 0:
                logger.info('Computing traces for %s run...', str(self))

            for parameters in self.monitor.parameters:

                self.policy.parameters = parameters

                # compute trace
                trace = self.environment._rollout(self.policy)
                self.monitor.traces.append(trace)

                # compute total reward
                reward = sum([t[2] for t in trace])
                self.monitor.rewards.append(reward)

    def _before_step(self):
        """Monitor algorithm before step.

        Parameters
        ----------
        alg :
            Algorithm instance to be monitored.
        """
        # count the number of rollouts for each step
        self.environment.monitor.rollout_cnt = 0

        if config.monitor_verbosity > 2:
            logger.info('Computing step %d for %s...', self.monitor.step_cnt,
                        str(self))

    def _after_step(self):
        """Monitor algorithm after step.

        Parameters
        ----------
        alg :
            Algorithm instance to be monitored.
        """
        emonitor = self.environment.monitor

        self.monitor.step_cnt += 1

        # store the number of rollouts
        self.monitor.rollout_cnts.append(emonitor.rollout_cnt)

        # retrieve information from the policy
        if self.has_policy:
            # retrieve current parameters
            parameters = self.policy.parameters
            # store information
            self.monitor.parameters.append(parameters)

        # log if wanted
        self._step_log()

    def _step_log(self):
        # print information if wanted
        monitor = self.monitor
        n = monitor.step_cnt
        log = 0

        # check verbosity level
        if config.monitor_verbosity > 0:
            if monitor.step_cnt % 1000 == 0:
                log = 1000

        if config.monitor_verbosity > 1:
            if monitor.step_cnt % 100 == 0:
                log = 100

        if config.monitor_verbosity > 2:
            log = 1

        if log:
            # generate time strings
            now = time.time()
            t = now - monitor.optimize_start
            t_s = "{:.2f}".format(t)
            avg_s = "{:.3f}".format(t / n)

            # generate log message
            msg = 'Status for ' + self.__class__.__name__ + ' on '
            msg += self.environment.__class__.__name__ + ':\n\n'
            msg += '\tRun: %d\tTime: %s\t Avg: %s\n' % (n, t_s, avg_s)
            if self.has_policy:
                # retrieve current state
                par_s = str(self.policy.parameters)
                msg += '\tParameter: \t%s\n' % (par_s)

            logger.info(msg)


class EnvData(object):
    """Class to store environment tracking data.

    Attributes
    ----------
    rollout_cnt : Int
        number of rollouts performed on environment.
    """

    def __init__(self):
        """Initialize attributes."""
        self.rollout_cnt = 0


class AlgoData(object):
    """Class used to store algorithm tracking data.

    Attributes
    ----------
    optimize_start : Float
        Start time of the optimization.
    optimize_time : Float
        Start time of intermediate runs.
    step_cnt : Int
        Number of steps performed since initialization.
    rollout_cnts : List
        Number of rollouts during one step.
    parameters : List
        List of parameters found during optimization.
    traces : List
        List of traces for parameters.
    rewards : List
        List of rewards for parameters.
    """

    def __init__(self):
        """Initialize attributes."""
        self.reset()

    def reset(self):
        """Reset monitor data."""
        self.optimize_start = 0
        self.optimize_time = 0

        self.step_cnt = 0
        self.rollout_cnts = []

        self.parameters = []
        self.traces = []
        self.rewards = []



"""Tests for spaces module."""
from __future__ import absolute_import

from functools import partial
import inspect

from numpy import array
import SafeRLBench.spaces as spaces


"""Dictionary storing initialization arguments for classes."""
class_arguments = {
    spaces.BoundedSpace: [array([-1, -2]), array([1, 0])],
    spaces.RdSpace: [(3, 2)],
    spaces.DiscreteSpace: [5]
}


class TestSpaces(object):
    """Wrap spaces tests."""

    classes = []

    @classmethod
    def setUpClass(cls):
        """Initialize classes list."""
        for name, c in inspect.getmembers(spaces):
            if inspect.isclass(c):
                cls.classes.append(c)

    def exhaustive_tests(self):
        """Check: Spaces tests initial values for testing."""
        for c in self.classes:
            if c not in class_arguments:
                assert(False)

    def generate_tests(self):
        """Generate tests for spaces implementations."""
        for c in self.classes:
            if c in class_arguments:
                check = partial(self.check_contains)
                check.description = ('Test: ' + c.__name__.upper()
                                     + ': implementation.')
                yield check, c

    def check_contains(self, c):
        """Check if contains and element is implemented."""
        space = c(*class_arguments[c])
        try:
            x = space.sample()
            b = space.contains(x)
        except NotImplementedError:
            assert(False)
        assert(b)



"""Exceptions and error messages."""

import logging

logger = logging.getLogger(__name__)


class NotSupportedException(Exception):
    """Exception raised when requirements are not installed.

    Attributes
    ----------
    dep : Module
        The dependent module.
    dep_name : String
        Name of the dependency for a meaningful error message.
    """

    def __init__(self, dep, name='Some'):
        """Initialize NotSupportedException.

        Parameters
        ----------
        dep : Module
            The dependent module.
        dep_name : String
            Name of the dependency for a meaningful error message.
        """
        msg = name + " is not installed on this system."

        super(NotSupportedException, self).__init__(msg)

        self.dep = dep
        self.name = name


class MultipleCallsException(Exception):
    """Exception raised when a setup method is called multiple times."""

    pass


class IncompatibilityException(Exception):
    """Exception raised when any two parts are incompatible with each other.

    Attributes
    ----------
    obj1 : object
        Instance of the object calling the exception.
    obj2 : object
        Instance of the object being incompatible.
    """

    def __init__(self, obj1, obj2):
        """Initialize IncompatibilityException.

        Parameters
        ----------
        obj1 : object
            Instance of the object calling the exception.
        obj2 : object
            Instance of the object being incompatible.
        """
        msg = "%s is incompatible with %s." % (self.obj2.__name__,
                                               self.obj1.__name__)

        super(IncompatibilityException, self).__init__(msg)

        self.obj1 = obj1
        self.obj2 = obj2


def add_dependency(dep, dep_name='Some'):
    """Add dependency.

    Function, that will raise a `NotSupportedException` when `dep` is None.

    Parameters
    ----------
    dep : Module
        The dependent module.
    dep_name : String
        Name of the dependency for a meaningful error message.
    """
    if dep is None:
        raise NotSupportedException(dep, dep_name)



"""Bounded subspace of R^n."""
import numpy as np
from SafeRLBench import Space

from numpy.random import rand


class BoundedSpace(Space):
    """Bounded subspace of R^n.

    Attributes
    ----------
    lower : array-like
        Lower bound
    upper : array-like
        Upper bound

    Examples
    --------
    The `BoundedSpace` class can be instatiated in two ways. If you have
    individual bounds for each dimension, then you can directly pass the
    `lower` or `upper` bound as an array-like.

    >>> space = BoundedSpace(np.array([-1, -2]), np.array([1, 0]))

    In this case the shape argument will be ignored. If you want to create a
    however shaped box, where all the bounds are the same, then, you may pass
    a lower and an upper bound as a scalar and make sure that you specify the
    shape.

    >>> space = BoundedSpace(-1, 1, shape=(2,))
    """

    def __init__(self, lower, upper, shape=None):
        """Initialize BoundedSpace.

        Parameters
        ----------
        lower : array-like
            Lower bound of the space. Either an array or an integer.
            Must agree with the input of the upper bound.
        upper : array-like
            Upper bound of the space. Either an array or an integer. Must
            agree with the input of the lower bound.
        shape : integer
            Shape of the bounds. Input will be ignored, if the bounds are non
            scalar, if they are scalar, it must be set.
        """
        if (np.isscalar(lower) and np.isscalar(upper)):
            assert shape is not None, "Shape must be set, if bounds are scalar"
            self.lower = np.zeros(shape) + lower
            self.upper = np.zeros(shape) + upper
        else:
            self.lower = np.array(lower)
            self.upper = np.array(upper)
            assert self.lower.shape == self.upper.shape, "Shapes do not agree."

        self._dim = None

    def contains(self, x):
        """Check if element is contained."""
        return (x.shape == self.lower.shape
                and (x >= self.lower).all()
                and (x <= self.upper).all())

    def sample(self):
        """Return element."""
        element = rand(*self.shape) * (self.upper - self.lower) + self.lower
        return element

    @property
    def shape(self):
        """Return element shape."""
        return self.lower.shape

    @property
    def dimension(self):
        """Return dimension of the space."""
        if self._dim is None:
            d = 1
            for i in range(len(self.shape)):
                d *= self.shape[i]
            self._dim = d
        return self._dim

    def __repr__(self):
        return 'BoundedSpace(lower=%s, upper=%s)' % (str(self.lower),
                                                     str(self.upper))



"""Discrete space implementation."""

from SafeRLBench import Space

import numpy as np


class DiscreteSpace(Space):
    """Discrete Space.

    Let d be the dimension of the space, then it will contain elements
    {0, 1, ... , dim-1}.

    Examples
    --------
    Create a `DiscreteSpace` with three states:
    >>> from SafeRLBench.spaces import DiscreteSpace
    >>> discrete_space = DiscreteSpace(3)
    """

    def __init__(self, dim):
        """Initialize `DiscreteSpace`.

        Parameters
        ----------
        dim : int
            Number of states.
        """
        assert dim > 0, ("If you need a discrete space without elements, you "
                         + "do not need this class.")
        self._dim = dim

    def contains(self, x):
        """Check if element is part of the space."""
        return (isinstance(x, int) and x >= 0 and x < self._dim)

    def sample(self):
        """Sample an element of the space."""
        return np.random.randint(self._dim)

    @property
    def dimension(self):
        """Return dimension of the space."""
        return self._dim

    def __repr__(self):
        return 'DiscreteSpace(dim=%d)' % self._dim



"""Global Configuration Class."""
import logging
import sys


class SRBConfig(object):
    """SafeRLBench configuration class.

    This is a configuration class providing a container for global variables
    and configuration functions.

    In general this class should not be instantiated directly, but rather
    accessed through the global variable ``SafeRLBench.config``, which is
    created when the package is imported and will contain the root logger of
    the package.

    Attributes
    ----------
    logger_stream_handler :
        This is a property wrapping the current stream handler. The current
        stream handler can be accessed through this property, or it may even
        be replaced with a new stream handler. In case of resetting the stream
        handler, the old handler will be removed from the logger
        automatically.
    logger_file_handler :
        This is a property wrapping the current file handler. The current
        file handler can be accessed through this property, or it may even
        be replaced with a new stream handler. In case of resetting the file
        handler, the old handler will be removed from the logger
        automatically.
    logger_format :
        This is a property to access the format stored. This is the default
        format that will be used when adding the default handlers.
        When assigned to, the formats of already set loggers will be changed
        to the new format.
    log :
        The logger object.
    n_jobs :
        Number of jobs used by the library
    monitor_verbosity :
        Verbosity of the monitor.

    Methods
    -------
    monitor_set_verbosity(verbosity)
        Set monitor verbosity level.
    jobs_set(n_jobs)
        Set the amount of jobs used by a worker pool.
    logger_set_level(level=logging.INFO)
        Set the logger level package wide.
    logger_add_stream_handler()
        Set a handler to print logs to stdout.
    logger_add_file_handler(path)
        Set a handler to print to file.

    Notes
    -----
    Access logger levels through the static variables:

    +-----------+------------------+
    |DEBUG      | logging.DEBUG    |
    +-----------+------------------+
    |INFO       | logging.INFO     |
    +-----------+------------------+
    |WARNING    | logging.WARNING  |
    +-----------+------------------+
    |ERROR      | logging.ERROR    |
    +-----------+------------------+
    |CRITICAL   | logging.CRITICAL |
    +-----------+------------------+
    """

    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL

    def __init__(self, log):
        """Initialize default configuration."""
        # some libraries think it is a good idea to add handlers by default
        # without documenting that at all, thanks gpy...
        log.propagate = False

        self.log = log
        self.n_jobs = 1
        self.monitor_verbosity = 0

        self._stream_handler = None
        self._file_handler = None
        self._fmt = ('%(process)d - %(asctime)s - %(name)s - %(levelname)s'
                     + ' - %(message)s')
        self._formatter = logging.Formatter(self._fmt)

        try:
            import gym
            gym.undo_logger_setup()
        except:
            pass

    def monitor_set_verbosity(self, verbosity):
        """Set monitor verbosity level.

        Parameters
        ----------
        verbose : int
            Non negative verbosity level
        """
        if verbosity < 0:
            raise ValueError('Verbosity level can not be negative.')
        self.monitor_verbosity = verbosity

    def jobs_set(self, n_jobs):
        """Set the amount of jobs used by a worker pool.

        Parameters
        ----------
        n_jobs : Int
            Number of jobs, needs to be larger than 0.
        """
        if n_jobs <= 0:
            raise ValueError('Number of jobs needs to be larger than 0.')
        self.n_jobs = n_jobs

    def logger_set_level(self, level=logging.INFO):
        """Set the logger level package wide.

        Parameters
        ----------
        level :
            Logger level as defined in logging.
        """
        self.log.setLevel(level)

    @property
    def logger_stream_handler(self):
        """Property storing the current stream handler.

        If overwritten with a new stream handler, the logger will be updated
        with the new stream handler.

        Examples
        --------
        Setup a stream handler for the logger.

        >>> from SafeRLBench import config
        >>> import logging
        >>> # configurate stream handler
        >>> ch = logging.StreamHandler(sys.stdout)
        >>> config.logger_stream_handler = ch

        To use the default format:

        >>> formatter = logging.Formatter(config.logger_format)
        >>> ch.setFormatter(formatter)

        which is equivalent to using `logger_add_stream_handler`.
        """
        return self._stream_handler

    @logger_stream_handler.setter
    def logger_stream_handler(self, ch):
        """Setter method for logger_stream_handler property."""
        if self._stream_handler is not None:
            self.log.removeHandler(self._stream_handler)

        self._stream_handler = ch
        if ch is not None:
            self.log.addHandler(ch)

    @property
    def logger_file_handler(self):
        """Property storing the current file handler.

        If overwritten with a new file handler, the logger will be updated with
        the new file handler.

        Examples
        --------
        Setup a stream handler for the logger.

        >>> from SafeRLBench import config
        >>> import logging
        >>> # configurate stream handler
        >>> fh = logging.FileHandler('logs.log')
        >>> config.logger_file_handler = fh

        To use the default format:

        >>> formatter = logging.Formatter(config.logger_format)
        >>> fh.setFormatter(formatter)

        which is equivalent to using `logger_add_file_handler`.
        """
        return self._file_handler

    @logger_file_handler.setter
    def logger_file_handler(self, fh):
        """Setter method for logger_file_handler property."""
        if self._file_handler is not None:
            self.log.removeHandler(self._file_handler)

        self._file_handler = fh
        if fh is not None:
            self.log.addHandler(fh)

    @property
    def logger_format(self):
        """Property for default logger format.

        If overwritten stream and file handler will be updated accordingly.
        However if manually updating stream and file handler logger_format will
        be ignored.
        """
        return self._fmt

    @logger_format.setter
    def logger_format(self, fmt):
        """Setter method for logger_format property."""
        self._formatter = logging.Formatter(fmt)

        self._fmt = fmt

        if self.logger_stream_handler is not None:
            self.logger_stream_handler.setFormatter(self._formatter)

        if self.logger_file_handler is not None:
            self.logger_file_handler.setFormatter(self._formatter)

    def logger_add_stream_handler(self):
        """Set a handler to print logs to stdout."""
        if self._stream_handler is not None:
            self.log.removeHandler(self._stream_handler)

        ch = logging.StreamHandler(sys.stdout)
        ch.setFormatter(self._formatter)

        self._stream_handler = ch
        self.log.addHandler(ch)

    def logger_add_file_handler(self, path):
        """Set a handler to print to file.

        Parameters
        ----------
        path :
            Path to log file.
        """
        if self._file_handler is not None:
            self.log.removeHandler(self.file_handler)

        fh = logging.FileHandler(path)
        fh.setFormatter(self._formatter)

        self._file_handler = fh
        self.log.addHandler(fh)



"""Linear Car."""
import numpy as np
from numpy import copy, array
from numpy.linalg import norm

from SafeRLBench import EnvironmentBase
from SafeRLBench.spaces import RdSpace, BoundedSpace


# TODO: LinearCar: add examples
class LinearCar(EnvironmentBase):
    """Implementation of LinearCar Environment.

    This is a very simple environment implementing a car in an arbitrarily
    dimensioned space. By default it will just be one dimensional, which
    results in a two dimensional state space, that is, (pos, vel), and
    accordingly in a one dimensional bounded action space, that is, the
    acceleration.

    Attributes
    ----------
    state : ndarray
        Current state of the LinearCar.
    initial_state : ndarray
        Initial state of the LinearCar.
    goal : ndarray
        Goal state.
    eps : float
        Margin for completion. If 0, the goal is to stabilize at the goal
        completly.
    step : float
        Update step.
    state_space : Space object
        State space as deduced from the state.
    action_space : Space object
        Action space as deduced from the state.
    """

    def __init__(self, state=array([[0.], [0.]]), goal=array([[1.], [0.]]),
                 step=0.01, eps=0, horizon=100):
        """
        Initialize LinearCar.

        Parameters
        ----------
        state : ndarray
            Initial state of the LinearCar. The state and action space will be
            deduced from this. The shape needs to be (2, d) for d > 0.
        goal : ndarray
            Goal state of the LinearCar. The shape should comply to the shape
            of the initial state.
            In case the velocity is non-zero, eps should be strictly greater
            than zero, since there is no way for the system to stabilize in
            the goal state anyway.
        eps : float
            Reward at which we want to abort. If zero we do not abort at all.
        step : float
            Update step.
        """
        assert state.shape[0] == 2, 'Invalid shape of the initial state.'
        assert state.shape == goal.shape, 'State and goal shape have to agree.'

        # Initialize EnivronmentBase attributes
        self.horizon = horizon
        self.state_space = RdSpace(state.shape)
        self.action_space = BoundedSpace(-1, 1, shape=(state.shape[1],))

        # Initialize State
        self.initial_state = state
        self.state = copy(state)

        # Initialize Environment Parameters
        self.goal = goal
        self.eps = eps
        self.step = step

    def _update(self, action):
        one = np.ones(self.action_space.shape)
        action = np.maximum(np.minimum(action, one), -one)

        self.state[1] += self.step * action
        self.state[0] += self.state[1]

        return (action, copy(self.state), self._reward())

    def _reset(self):
        self.state = copy(self.initial_state)

    def _rollout(self, policy):
        self.reset()
        trace = []
        for n in range(self.horizon):
            action = policy(self.state)
            trace.append(self.update(action))
            if (self.eps != 0 and self._achieved()):
                return trace
        return trace

    def _reward(self):
        return -norm(self.state - self.goal)

    def _achieved(self):
        return (abs(self._reward()) < self.eps)



"""Algorithm Module.

=================== =========================================
Algorithm
=============================================================
A3C                 Asynchronous Actor-Critic Agents
PolicyGradient      Different Policy Gradient Implementations
DiscreteQLearning   Q-Learning using a table
SafeOpt             Bayesian Optimization with SafeOpt
SafeOptSwarm        Bayesion Optimization with SafeOptSwarm
=================== =========================================
"""

from .policygradient import PolicyGradient
from .safeopt import SafeOpt, SafeOptSwarm
from .a3c import A3C
from .q_learning import DiscreteQLearning

__all__ = ['PolicyGradient', 'SafeOpt', 'A3C', 'DiscreteQLearning',
           'SafeOptSwarm']



from __future__ import absolute_import

import logging

from .configuration import SRBConfig

# Initialize configuration
config = SRBConfig(logging.getLogger(__name__))

from .monitor import AlgoMonitor, EnvMonitor
from .base import EnvironmentBase, Space, AlgorithmBase, Policy, ProbPolicy
from .bench import Bench, BenchConfig
from . import algo
from . import envs
from . import policy
from . import spaces
from . import error
from . import measure

# Add things to all
__all__ = ['EnvironmentBase',
           'Space',
           'AlgorithmBase',
           'Policy',
           'ProbPolicy',
           'AlgoMonitor',
           'EnvMonitor',
           'SRBConfig',
           'Bench',
           'BenchConfig',
           'envs',
           'algo',
           'policy',
           'spaces',
           'measure',
           'error']


# Import test after __all__ (no documentation)
# from numpy.testing import Tester
# test = Tester().test



from __future__ import absolute_import

from .general_mountaincar import GeneralMountainCar
from .linear_car import LinearCar
from .gym_wrap import GymWrap
from .quadrocopter import Quadrocopter
from .mdp import MDP

__all__ = [
    'GeneralMountainCar',
    'LinearCar',
    'GymWrap',
    'Quadrocopter',
    'MDP'
]

# TODO: Envs: Add module docs in __init__ file.



"""General Mountain Car."""
import numpy as np
from numpy import pi, array, copy, cos, sin

from SafeRLBench.base import EnvironmentBase
from SafeRLBench.spaces import BoundedSpace


class GeneralMountainCar(EnvironmentBase):
    """Implementation of a GeneralMountainCar Environment.

    Attributes
    ----------
    state_space : BoundedSpace
        Space object describing the state space.
    action_space : BoundedSpace
        Space object describing the action space.
    state : array-like
        Current state of the car.
    initial_state : array-like
        Initial state of the car.
    gravitation : double
    power : double
    goal : double
        Goal along x-coordinate
    """

    def __init__(self,
                 state_space=BoundedSpace(array([-1, -0.07]),
                                          array([1, 0.07])),
                 action_space=BoundedSpace(-1, 1, shape=(1,)),
                 state=np.array([0, 0]),
                 contour=None, gravitation=0.0025, power=0.0015,
                 goal=0.6, horizon=100):
        """Initialize GeneralMountainCar Environment.

        Parameters
        ----------
        state_space : BoundedSpace
            Space object describing the state space.
        action_space : BoundedSpace
            Space object describing the action space.
        state : array-like
            Initial state of the car
        contour : tuple of callables
            If contour is None, a default shape will be generated. A valid
            tuple needs to contain a function for the height at a position
            in the first element and a function for the gradient at a position
            in the second argument.
        gravitation : double
        power : double
        goal : double
            Goal along x-coordinate
        """
        # Initialize Environment Base Parameters
        super(GeneralMountainCar, self).__init__(state_space,
                                                 action_space,
                                                 horizon)

        # setup environment parameters
        self.goal = goal
        self.power = power
        self.gravitation = gravitation

        # setup contour
        if contour is None:
            def _hx(x):
                return -cos(pi * x)
            self._hx = _hx

            def _dydx(x):
                return pi * sin(pi * x)
            self._dydx = _dydx
        else:
            self._hx = contour[0]
            self._dydx = contour[1]

        # init state
        self.state = copy(state)
        self.initial_state = state

    def _update(self, action):
        """Compute step considering the action."""
        action = array(action).flatten()
        action = max(min(action, 1.0), -1.0)

        if hasattr(action, 'size') and action.size == 1:
            action_in = action[0]
        else:
            action_in = action

        position = self.state[0]
        velocity = self.state[1]

        velocity += (action_in * self.power
                     - self._dydx(position) * self.gravitation)
        position += velocity

        bounds = self.state_space

        velocity = max(min(velocity, bounds.upper[1]), bounds.lower[1])
        position = max(min(position, bounds.upper[0]), bounds.lower[0])

        # make sure outputs have the right form
        self.state = np.array([position, velocity])
        action = np.reshape(action, self.action_space.shape)

        return action, copy(self.state), self._reward()

    def _reset(self):
        self.state = copy(self.initial_state)

    def _reward(self):
        return(self.height() - 1)

    def _rollout(self, policy):
        self.reset()
        trace = []
        for n in range(self.horizon):
            action = policy(self.state)
            trace.append(self.update(action))
            if (self.position() >= self.goal):
                return trace
        return trace

    def height(self):
        """Compute current height."""
        return(self._hx(self.state[0].item()).item())

    def position(self):
        """Compute current position in x."""
        return(self.state[0])



"""Quadrocopter environment wrapper."""

from __future__ import division, print_function, absolute_import

from SafeRLBench import EnvironmentBase
from SafeRLBench.spaces import RdSpace

from ._quadrocopter import QuadrotorDynamics
from ._quadrocopter import StateVector

from functools import partial

from six import string_types

import numpy as np
from numpy import array
from numpy import pi, cos, sin
from numpy.linalg import norm

import logging

logger = logging.getLogger(__name__)

# Available reference functions.
REFERENCE_TYPES = ['circle', 'stationary', 'oscillate']


class Quadrocopter(EnvironmentBase):
    """Quadrocopter simulation.

    Attributes
    ----------
    horizon : int
        Number of iterations for the main simulation
    pre_sim_horizon : int
        Number of iterations for the pre-simulation.
    _model : model object
        Object simulating the quadrotor dynamics.
    """

    def __init__(self,
                 init_pos=None, init_vel=None, num_sec=9,
                 num_init_sec=4, ref='circle', period=1 / 70.,
                 seed=None):
        """Quadrocopter initialization.

        Parameters
        ----------
        init_pos : array-like
            Initial position of the quadrocopter. Default: None ; which will
            set init_pos to [1, 0, 0].
        init_vel : array-like
            Initial velocity of the quadrocopter. Default: None ; which will
            set init_vel to [0, pi / 2, 0]
        num_sec : integer
        num_init_sec : integer
        ref : string or reference object
            Name of the reference. Currently supported are 'circle',
            'stationary' or 'oscillate'.
        period : float
        seed : int
        """
        # spaces
        self.state_space = RdSpace((22,))
        self.action_space = RdSpace((4,))

        # seed
        if seed is not None:
            np.random.seed = seed
            self._seed = seed

        # initial position
        if init_pos is None:
            init_pos = array([cos(0), sin(0), 0.])

        if len(init_pos) != 3:
            raise ValueError("init_pos with invalid length %d.", init_pos)

        # initial velocity
        if init_vel is None:
            init_vel = array([-pi / 2. * sin(0), pi / 2. * cos(0), 0.])

        if len(init_vel) != 3:
            raise ValueError("init_vel with invalid length %d.", init_vel)

        # initialize model
        self._model = QuadrotorDynamics(init_pos, init_vel)

        if isinstance(ref, string_types):
            self.reference = Reference(ref, period)
        else:
            self.reference = ref
            self.period = ref.period

        self.reference.reset(self.state)

        self.horizon = int(1. / period) * num_sec
        self.pre_sim_horizon = int(1. / period) * num_init_sec

        self.period = self.reference.period

        self._init_pos = init_pos
        self._init_vel = init_vel

        self._trajectory = np.atleast_2d(np.zeros(3))
        self._time = []
        self._step = 0

    def _update(self, action):
        assert self.action_space.contains(action), "Invalid action."

        self._model.update_position(action)

        self._step += 1
        time = self._step * self.period

        self._time.append(time)
        self._trajectory = np.vstack((self._trajectory, self.state.pos))

        reward = self._reward()
        self.reference.update(self.state, time)

        return action, self.state.copy(), reward

    def _reset(self):
        self._model = QuadrotorDynamics(self._init_pos, self._init_vel)
        self.reference.reset(self.state)
        self._trajectory = np.atleast_2d(np.zeros(3))
        self._time = []
        self._step = 0

    def _rollout(self, policy):
        if hasattr(policy, 'reference'):
            policy.reference = self.reference
        self.reset()
        trace = []
        for n in range(self.horizon):
            action = policy(self.state)
            trace.append(self.update(action))
        return trace

    def _reward(self):
        state = self.state
        ref = self.reference.reference

        reward = -norm(state.pos - ref.pos) - norm(state.vel - ref.vel)

        if np.isnan(reward):
            reward = -1.79769313e+308

        return reward

    @property
    def seed(self):
        """Seed."""
        return self._seed

    @seed.setter
    def seed(self, value):
        np.random.seed(value)
        self._seed = value

    @property
    def state(self):
        """Provide access to state_vector."""
        # this whole state vector implementation is annoyingly inefficient.
        return self._model.state.state_vector

    @state.setter
    def state(self, state):
        self._model.state.state_vector = state.view(StateVector)


class Reference(object):
    """Reference object for quadrocopter environment."""

    def __init__(self, name='circle', period=1. / 70, keep_record=True,
                 **kwargs):
        """Initialize Reference.

        Parameters
        ----------
        name : str
            The name of the reference function.
        period : float
            The time step that the simulation takes every iteration.
        keep_record : bool
            Whether the history of the reference object should be saved.
        **kwargs : dict
        """
        # name type checking.
        if not isinstance(name, string_types):
            raise ValueError('Invalid type for argument name.')
        if name not in REFERENCE_TYPES:
            raise ValueError(name + ' is not a valid reference.')

        self._name = name
        self.period = period
        self._iter = 0
        self._reference_function = self._reference_chooser(**kwargs)
        self._current_ref = None
        self.keep_record = keep_record
        if keep_record:
            self._record = []

    @property
    def name(self):
        """Return the type of reference function."""
        return self._name

    @name.setter
    def name(self, value):
        if value not in REFERENCE_TYPES:
            raise ValueError(value + ' is not a valid reference.')

        self.reset()

        self._name = value
        self._reference_function = self._ref_chooser()

    @property
    def record(self):
        """Return the reference record of the simulation."""
        if self.keep_record:
            return np.atleast_2d(self._record)
        else:
            logger.warning("Reference record has not been saved.")

    def reset(self, state=None):
        """Reset internal state."""
        self._iter = 0
        self._current_ref = self._reference_function(state, 0, False)
        if self.keep_record:
            self._record = []

    def update(self, state, time, finished=False):
        """Compute the state of the reference object."""
        ref = self._reference_function(state, time, finished)
        self._iter += 1

        if self.keep_record:
            ref_value = np.hstack((ref.pos, ref.vel, ref.euler, ref.omega_b))
            self._update_record(ref_value)

        self._current_ref = ref

        return ref

    @property
    def reference(self):
        """Return the reference."""
        return self._current_ref

    def _update_record(self, ref_value):
        self._record.append(ref_value)
        assert self._iter == len(self._record)

    def _reference_chooser(self, **kwargs):
        # CIRCLE
        if self._name == 'circle':
            if kwargs.get('speed', False):
                speed = kwargs['speed']
            else:
                speed = pi / 2.
            if kwargs.get('initial_angle', False):
                init_angle = kwargs['initial_angle']
            else:
                init_angle = 0.
            if kwargs.get('radius', False):
                radius = kwargs['radius']
            else:
                radius = 1.
            if kwargs.get('z_vel', False):
                z_vel = kwargs['z_vel']
            else:
                z_vel = 0.
            return partial(_circle_reference,
                           speed=speed,
                           init_angle=init_angle,
                           radius=radius,
                           z_vel=z_vel)
        # STATIONARY
        elif self._name == 'stationary':
            if kwargs.get('position', False):
                position = kwargs['position']
            else:
                position = [1., 0., 0.]
            return partial(_stationary_reference,
                           position=position)
        # OSCILLATE
        elif self._name == 'oscillate':
            if kwargs.get('x_velocity', False):
                x_vel = kwargs['x_velocity']
            else:
                x_vel = 0.5
            if kwargs.get('omega', False):
                omega = kwargs['omega']
            else:
                omega = 1.
            if kwargs.get('radius', False):
                radius = kwargs['radius']
            else:
                radius = 0.5
            return partial(_oscillate_reference,
                           x_vel=x_vel,
                           omega=omega,
                           radius=radius)


# private circle reference function
def _circle_reference(state,
                      time,
                      finished,
                      radius=None,
                      speed=None,
                      init_angle=None,
                      z_vel=None):
    ref = StateVector()
    angle = init_angle + speed / radius * time

    ref.pos = array([radius * cos(angle),
                     radius * sin(angle),
                     z_vel * time])
    ref.vel[:] = [-speed * sin(angle), speed * cos(angle), z_vel]
    ref.euler[2] = pi + np.arctan2(state.pos[1], state.pos[0])
    # reference.omega_b[2] = speed / radius
    return ref


# private stationary reference function
def _stationary_reference(state,
                          time,
                          finished,
                          position=None):
    ref = StateVector()
    ref.pos[0] = position[0]
    ref.pos[1] = position[1]
    ref.pos[2] = position[2]
    return ref


# private oscillation reference function
def _oscillate_reference(state,
                         time,
                         finished,
                         x_vel=None,
                         omega=None,
                         radius=None):
    ref = StateVector()
    angle = omega * time
    ref.pos[0] = x_vel * time
    ref.pos[1] = radius * sin(angle)
    ref.pos[2] = 0.
    ref.vel[0] = x_vel
    ref.vel[1] = radius * omega * cos(angle)
    ref.vel[2] = 0.
    return ref



"""Q-learning implementations."""

from SafeRLBench import AlgorithmBase, Policy
from SafeRLBench.spaces import DiscreteSpace
from SafeRLBench.error import IncompatibilityException

import numpy as np


# TODO: DiscreteQLearning: examples, monitoring, finished, adaptive rate
class DiscreteQLearning(AlgorithmBase):
    """Q-Learning Algorithm.

    This Algorithm estimates a quality measure that maps every (state, action)
    pair to a real number.

    Attributes
    ----------
    Q : ndarray
        Array representing the quality for each state action pair.
    environment :
        The environment for which we want to estimate the Q function. Its
        state and action space need to be an instance of `DiscreteSpace`.
    discount : float
        Discount factor.
    max_it : int
        Maximum number of iterations.
    rate : float
        Update rate.
    shape : (int, int)
        Tuple containing the dimension of the state and action space.

    Notes
    -----
    The environment needs to use a discrete state and action space, because
    this Q-Learning implementation uses a table to estimate the Q function.
    """

    def __init__(self, environment, discount, max_it, rate):
        """Initialize QLearning.

        Parameters
        ----------
        environment :
            The environment for which we want to estimate the Q function. Its
            state and action space need to be an instance of `DiscreteSpace`.
        discount : float
            Discount factor.
        max_it : int
            Maximum number of iterations.
        rate : float
            Update rate.
        """
        # make some sanity checks
        if (not isinstance(self.environment.action_space, DiscreteSpace)
           and not isinstance(self.environment.state_space, DiscreteSpace)):
            raise IncompatibilityException(self, self.environment)

        if discount <= 0:
            raise ValueError('discount %d needs to be larger than zero.',
                             discount)

        if max_it <= 0:
            raise ValueError('max_it %d needs to be larger than zero.', max_it)

        # initialize the fields
        self.environment = environment
        self.discount = discount
        self.max_it = max_it
        self.rate = rate

        # determine the dimension of the state and action space
        d_state = environment.state_space.dimension
        d_action = environment.action_space.dimension

        self.shape = (d_state, d_action)

        # initialize the lookup table for the Q function.
        self.Q = None
        self.policy = _RandomPolicy(environment.action_space)

    def _initialize(self):
        self.Q = np.zeros(self.shape)

    def _step(self):
        trace = self.environment.rollout(self.policy)
        for (action, state, reward) in trace:
            dq = (reward + self.discount * self.Q[state, :].max()
                  - self.Q[state, action])
            self.Q[state, action] += self.rate * dq

    def _is_finished(self):
        pass

    # TODO: Q-learning evaluate qlearning performance appropiately


class _RandomPolicy(Policy):

    def __init__(self, action_space):
        self.action_space = action_space

    def map(self, state):
        return self.action_space.sample

    @property
    def parameters(self):
        return self.action_space.dimension

    @property
    def parameter_space(self):
        return None



"""Policy Gradient implementations."""

from SafeRLBench import AlgorithmBase
from SafeRLBench.spaces import BoundedSpace

import numpy as np
from numpy.linalg import solve, norm

from abc import ABCMeta, abstractmethod
from six import add_metaclass

import logging

logger = logging.getLogger(__name__)


class PolicyGradient(AlgorithmBase):
    """Implementing many policy gradient methods.

    This uses standard gradient descent using different policy gradient
    estimators.

    Attributes
    ----------
    environment :
        Environment we want to optimize the policy on. This should be a
        subclass of `EnvironmentBase`.
    policy :
        Policy we want to find parameters for. This should be a subclass of
        `Policy`.
    estimator :
        Either an estimator object, that is a subclass of
        PolicyGradientEstimator or a string. A list of possible estimator
        strings can be found in the Notes section. By default 'reinforce' will
        be used
    eps : float
        The optimizer will stop optimization ones the norm of the gradient is
        smaller than `eps`.
    rate : float
        This is the rate we use for the updates in each step

    Notes
    -----
    These strings can be used to access the implemented estimators.

    +------------+---------------------------------+
    |'forward_fd'| Uses forward finite differences.|
    +------------+---------------------------------+
    |'central_fd'| Uses central finite differences.|
    +------------+---------------------------------+
    |'reinforce' | Classic reinforce estimator.    |
    +------------+---------------------------------+
    |'gpomdp'    | Uses GPOMDP estimator.          |
    +------------+---------------------------------+
    """

    def __init__(self,
                 environment, policy, estimator='reinforce',
                 max_it=1000, eps=0.0001, est_eps=0.001,
                 parameter_space=BoundedSpace(0, 1, (3,)),
                 rate=1, var=0.5):
        """Initialize PolicyGradient.

        Parameters
        ----------
        environment :
            Environment we want to optimize the policy on. This should be a
            subclass of `EnvironmentBase`.
        policy :
            Policy we want to find parameters for. This should be a subclass of
            `Policy`.
        estimator :
            Either an estimator object, that is a subclass of
            PolicyGradientEstimator or a string. A list of possible estimator
            strings can be found in the Notes section. By default 'reinforce'
            will be used
        eps : float
            The optimizer will stop optimization ones the norm of the gradient
            is smaller than `eps`.
        est_eps : float
            In case an estimator needs to converge, this is the margin it will
            use to stop.
        parameter_space :
        rate : float
            This is the rate we use for the updates in each step
        var : float
            This parameter will be used depending on the estimator type. e.g.
            for central differences this value corresponds to the grid size
            that is used.
        """
        super(PolicyGradient, self).__init__(environment, policy, max_it)

        self.parameter_space = policy.parameter_space

        self.eps = eps
        self.rate = rate

        if isinstance(estimator, str):
            estimator = estimators[estimator]
        elif issubclass(estimator, PolicyGradientEstimator):
            pass
        else:
            raise ImportError('Invalid Estimator')

        self.estimator = estimator(environment, self.parameter_space, max_it,
                                   est_eps, var)

    def _initialize(self):
        logger.debug("Initializing Policy.")
        # check if policy is already initialized by the user
        if self.policy.initialized:
            logger.debug("Use pre-set policy parameters.")
            return self.policy.parameters

        # outerwise draw an element at random from the parameter space
        parameter = self.parameter_space.sample()

        for _ in range(1000):
            self.policy.parameters = parameter
            grad = self.estimator(self.policy)

            if (norm(grad) >= 1000 * self.eps):
                return parameter

            parameter = self.parameter_space.sample()

        logger.error('Unable to find non-zero gradient.')

    def _step(self):
        grad = self.estimator(self.policy)

        parameter = self.policy.parameters

        self.policy.parameters = parameter + self.rate * grad

        self.grad = grad

    def _is_finished(self):
        done = False
        if np.isnan(self.grad).any():
            done = True
            logger.warning('Abort Optimization. Gradient contained not NaN')
        done = done or (norm(self.grad) < self.eps)
        return done


@add_metaclass(ABCMeta)
class PolicyGradientEstimator(object):
    """Interface for Gradient Estimators."""

    name = 'Policy Gradient'

    def __init__(self, environment, parameter_space, max_it=200, eps=0.001):
        """Initialize."""
        self.environment = environment
        self.state_dim = environment.state.shape[0]
        self.par_dim = parameter_space.dimension

        self.eps = eps
        self.max_it = max_it

    def __repr__(self):
        return self.__class__.__name__

    def __call__(self, policy):
        """Invoke _estimate_gradient(policy)."""
        return self._estimate_gradient(policy)

    @abstractmethod
    def _estimate_gradient(self, policy):
        pass


class ForwardFDEstimator(PolicyGradientEstimator):
    """Forward Finite Differences Gradient Estimator."""

    name = 'Forward Finite Differences'

    def __init__(self, environment, parameter_space=BoundedSpace(0, 1, (3,)),
                 max_it=200, eps=0.001, var=1):
        """Initialize."""
        super(ForwardFDEstimator, self).__init__(environment, parameter_space,
                                                 max_it, eps)
        self.var = var

    def _estimate_gradient(self, policy):
        env = self.environment
        var = self.var
        # store current policy parameter
        parameter = policy.parameters
        par_dim = policy.parameter_space.dimension

        # using forward differences
        trace = env.rollout(policy)
        j_ref = sum([x[2] for x in trace]) / len(trace)

        dj = np.zeros((2 * par_dim))
        dv = np.append(np.eye(par_dim), -np.eye(par_dim), axis=0)
        dv *= var

        for n in range(par_dim):
            variation = dv[n]

            policy.parameters = parameter + variation
            trace_n = env.rollout(policy)

            jn = sum([x[2] for x in trace]) / len(trace_n)

            dj[n] = j_ref - jn

        grad = solve(dv.T.dot(dv), dv.T.dot(dj))

        # reset current policy parameter
        policy.parameters = parameter

        return grad


class CentralFDEstimator(PolicyGradientEstimator):
    """Central Finite Differences Gradient Estimator."""

    name = 'Central Finite Differences'

    def __init__(self, environment, parameter_space=BoundedSpace(0, 1, (3,)),
                 max_it=200, eps=0.001, var=1):
        """Initialize."""
        super(CentralFDEstimator, self).__init__(environment, parameter_space,
                                                 max_it, eps)
        self.var = var

    def _estimate_gradient(self, policy):
        env = self.environment

        parameter = policy.parameters
        par_dim = policy.parameter_space.dimension

        dj = np.zeros((par_dim,))
        dv = np.eye(par_dim) * self.var / 2

        for n in range(par_dim):
            variation = dv[n]

            policy.parameters = parameter + variation
            trace_n = env.rollout(policy)

            policy.parameters = parameter - variation
            trace_n_ref = env.rollout(policy)

            jn = sum([x[2] for x in trace_n]) / len(trace_n)
            jn_ref = sum([x[2] for x in trace_n_ref]) / len(trace_n_ref)

            dj[n] = jn - jn_ref

        grad = solve(dv.T.dot(dv), dv.T.dot(dj))
        policy.parameters = parameter

        return grad


class ReinforceEstimator(PolicyGradientEstimator):
    """Reinforce Gradient Estimator."""

    name = 'Reinforce'

    def __init__(self, environment, parameter_space=BoundedSpace(0, 1, (3,)),
                 max_it=200, eps=0.001, lam=0.5):
        """Initialize."""
        super(ReinforceEstimator, self).__init__(environment, parameter_space,
                                                 max_it, eps)
        self.lam = lam

    def _estimate_gradient(self, policy):
        env = self.environment
        par_shape = policy.parameters.shape
        max_it = self.max_it

        b_div = np.zeros(par_shape)
        b_nom = np.zeros(par_shape)

        grads = np.zeros(par_shape)
        grad = np.zeros(par_shape)

        for n in range(max_it):
            trace = env.rollout(policy)

            lam = self.lam

            actions = [x[0] for x in trace]
            states = [x[1] for x in trace]

            rewards_sum = sum([x[2] * lam**k for k, x in enumerate(trace)])

            lg_sum = sum(list(map(policy.grad_log_prob, states, actions)))

            b_div_n = lg_sum**2
            b_nom_n = b_div_n * rewards_sum

            b_div += b_div_n
            b_nom += b_nom_n

            b = b_nom / b_div
            grad_n = lg_sum * (rewards_sum - b)

            grads += grad_n

            grad_old = grad
            grad = grads / (n + 1)

            if (n > 2 and norm(grad_old - grad) < self.eps):
                return grad

        logger.warning('ReinforceEstimator did not converge!'
                       + 'You may want to raise max_it.')
        return grad


class GPOMDPEstimator(PolicyGradientEstimator):
    """GPOMDP Gradient Estimator."""

    name = 'GPOMDP'

    def __init__(self, environment, parameter_space=BoundedSpace(0, 1, (3,)),
                 max_it=200, eps=0.001, lam=0.5):
        """Initialize."""
        super(GPOMDPEstimator, self).__init__(environment, parameter_space,
                                              max_it, eps)
        self.lam = lam

    def _estimate_gradient(self, policy):
        env = self.environment
        h = env.horizon
        shape = policy.parameters.shape

        b_nom = np.zeros((h, shape))
        b_div = np.zeros((h, shape))
        b = np.zeros((h, shape))
        grad = np.zeros(shape)

        lam = self.lam

        for n in range(self.max_it):
            trace = env.rollout(policy)
            b_n = np.zeros((h, shape))

            for k, state in enumerate(trace):
                update = policy.grad_log_prob(state[1], state[0])
                for j in range(k + 1):
                    b_n[j] += update

            fac = n / (n + 1)

            b_n = b_n**2
            b_div = fac * b_div + b_n / (n + 1)

            for k, state in enumerate(trace):
                b_nom[k] = fac * b_nom[k]
                b_nom[k] += b_n[k] * state[2] * lam**k / (n + 1)

            b = b_nom / b_div

            grad_update = np.zeros(shape)
            update = np.zeros(shape)
            for k, state in enumerate(trace):
                update += policy.grad_log_prob(state[1], state[0])
                grad_update += update * (-b[k] + state[2] * lam**k)

            if (n > 2 and norm(grad_update / (n + 1)) < self.eps):
                grad /= (n + 1)
                return grad
            grad += np.nan_to_num(grad_update)

        logger.warning('GPOMDP did not converge! '
                       + 'You may want to raise max_it.')
        grad /= n + 1
        return grad


"""Dictionary for resolving estimator strings."""
estimators = {
    'forward_fd': ForwardFDEstimator,
    'central_fd': CentralFDEstimator,
    'reinforce': ReinforceEstimator,
    'gpomdp': GPOMDPEstimator
}



"""Quadrocopter Controller."""
from SafeRLBench import Policy
from SafeRLBench.spaces import BoundedSpace
from SafeRLBench.envs._quadrocopter import StateVector

import numpy as np

import logging

logger = logging.getLogger(__name__)

__all__ = ('NonLinearQuadrocopterController')


# TODO: Controller: Documentation
class NonLinearQuadrocopterController(Policy):
    """Non-linear quadrocopter controller."""

    def __init__(self, zeta_z=0.7, params=[.7, .7, .7, .5, .707],
                 reference=None):
        """Initialize NonLinearQuadrocopterController.

        Parameters
        ----------
        zeta_z :
        params :
        reference :
        """
        self._zeta_z = zeta_z
        self._params = np.array(params)
        self.reference = reference

        if params is not None:
            self.initialized = True
        else:
            self.initialized = False

        self._par_space = BoundedSpace(np.array([0., 0., 0., 0., 0.]),
                                       np.array([1., 1., 1., 1., 1.]))

    def map(self, state):
        """Map state to action.

        Depends on a reference object. If the environment has a reference
        object it needs to set the reference at the start of the rollout.

        Parameters
        ----------
        state : array-like
            Element of state space.

        Returns
        -------
        action : ndarray
            Element of action space.
        """
        ref = self.reference.reference
        state = StateVector(state)

        # Allocate memory for the 4 outputs of the controller.
        action = np.empty((4,), dtype=np.float32)

        # Retrieve the different parameters and make sure the critical ones
        # are non zero.
        tau_x, tau_y, tau_z, tau_w, zeta = self._params
        if tau_x < 1e-3:
            tau_x = 1e-3
            logger.warning('Parameter `tau_x` too small for controller, '
                           + 'has been clipped to 1e-3"')
        if tau_y < 1e-3:
            tau_y = 1e-3
            logger.warning('Parameter `tau_y` too small for controller, '
                           + 'has been clipped to 1e-3"')
        if tau_w < 1e-3:
            tau_w = 1e-3
            logger.warning('Parameter `tau_w` too small for controller, '
                           + 'has been clipped to 1e-3"')
        if zeta < 1e-3:
            zeta = 1e-3
            logger.warning('Parameter `zeta` too small for controller, '
                           + 'has been clipped to 1e-3"')

        # desired acceleration in x and y (global coordinates, [m/s^2] )
        ax = (2. * zeta / tau_x * (ref.vel[0] - state.vel[0])
              + 1. / (tau_x**2) * (ref.pos[0] - state.pos[0]))
        ay = (2. * zeta / tau_y * (ref.vel[1] - state.vel[1])
              + 1. / (tau_y**2) * (ref.pos[1] - state.pos[1]))

        # Normalize by thrust
        thrust = np.linalg.norm(np.array([ax, ay, 9.81 + state.acc[2]]))
        ax /= thrust
        ay /= thrust

        # Rotate desired accelerations into the yaw-rotated inertial frame
        ax_b = ax * np.cos(state.euler[2]) + ay * np.sin(state.euler[2])
        ay_b = -ax * np.sin(state.euler[2]) + ay * np.cos(state.euler[2])

        # Get euler angles from rotation matrix
        action[1] = np.arcsin(-ay_b)
        action[0] = np.arcsin(ax_b / np.cos(action[1]))

        # Z-velocity command m/sec)
        action[2] = (2. * self._zeta_z / tau_z * (ref.vel[2] - state.vel[2])
                     + 1. / (tau_z**2) * (ref.pos[2] - state.pos[2]))

        # Yaw rate command (rad/sec)??
        yaw_err = (np.mod(ref.euler[2] - state.euler[2] + np.pi, 2 * np.pi)
                   - np.pi)
        action[3] = yaw_err / tau_w + ref.omega_b[2]

        return action

    @property
    def parameters(self):
        """Controller parameters."""
        return self._params

    @parameters.setter
    def parameters(self, params):
        self._params = np.array(params)

    @property
    def parameter_space(self):
        """Controller parameter space."""
        return self._par_space



"""Tests for envs module.

Need rework.
"""
from __future__ import absolute_import

# import unittest
# from numpy.testing import *
import inspect
from functools import partial

import SafeRLBench.envs as envs

import numpy as np

import gym
gym.undo_logger_setup()

from mock import Mock


class TestEnvironments(object):
    """
    Test Class for Environment tests.

    Note that you really dont want to inherit from unittest.TestCase here,
    because it will break reasonable output with verbose testing.
    """

    exclude = []

    args = {
        'GymWrap': envs.gym_wrap._get_test_args(),
        'MDP': envs.mdp._get_test_args()
    }

    @classmethod
    def setUpClass(cls):
        """Generate list of classes."""
        cls.classes = []
        for name, c in inspect.getmembers(envs):
            if inspect.isclass(c):
                cls.classes.append(c)

    def test_environment_requirements(self):
        """Generate tests for environment implementations."""
        for c in self.classes:
            if c.__name__ in self.exclude:
                pass
            else:
                # Generate NotImplementedError Test for _update
                check_update = partial(self.check_env_update)
                check_update.description = ('Test: ' + c.__name__.upper()
                                            + ': update implementation.')
                yield check_update, c

                # Generate NotImplementedError Test for _reset
                check_reset = partial(self.check_env_reset)
                check_reset.description = ('Test: ' + c.__name__.upper()
                                           + ': reset implementation.')
                yield check_reset, c

                check_rollout = partial(self.check_env_rollout)
                check_rollout.description = ('Test: ' + c.__name__.upper()
                                             + ': rollout implementation.')
                yield check_rollout, c

    def check_env_update(self, c):
        """Check if _update is implemented."""
        args = self.args.get(c.__name__, [])
        env = c(*args)
        x = env.action_space.sample()
        try:
            env._update(x)
        except NotImplementedError:
            assert False

    def check_env_reset(self, c):
        """Check if _reset is implemented."""
        args = self.args.get(c.__name__, [])
        env = c(*args)
        try:
            env._reset()
        except NotImplementedError:
            assert False

    def check_env_rollout(self, c):
        """Check rollout correctness at random positions."""
        args = self.args.get(c.__name__, [])
        env = c(*args)

        init_state = env.state

        def policy(state):
            return env.action_space.sample()

        policy_mock = Mock(side_effect=policy)
        trace = env._rollout(policy_mock)

        # reset the environment
        env._reset()
        env.state = init_state

        # if the environment depends on a seed, reset it.
        if hasattr(env, 'seed'):
            env.seed = env.seed

        actions = [t[0] for t in trace]

        policy_mock_redo = Mock(side_effect=actions)

        trace_verify = env._rollout(policy_mock_redo)

        for t, t_verify in zip(trace, trace_verify):
            print(t)
            print(t_verify)
            if isinstance(t[0], np.ndarray):
                assert(all(np.isclose(t_verify[0], t[0])))
            else:
                assert(np.isclose(t_verify[0], t[0]))
            if isinstance(t[1], np.ndarray):
                print(t_verify[1] - t[1])
                assert(all(np.isclose(t_verify[1], t[1])))
            else:
                assert(np.isclose(t_verify[1], t[1]))
            assert(np.isclose(t_verify[2], t[2]))



"""Policy tests."""
from __future__ import division, print_function, absolute_import

from SafeRLBench.spaces import BoundedSpace
from SafeRLBench.envs.quadrocopter import Reference
from SafeRLBench.envs._quadrocopter import StateVector
from SafeRLBench.policy import (NeuralNetwork,
                                LinearPolicy,
                                DiscreteLinearPolicy,
                                NonLinearQuadrocopterController)

import numpy as np
from numpy import isclose

import tensorflow as tf

from unittest2 import TestCase
from mock import Mock

import logging

logger = logging.getLogger(__name__)


class TestNeuralNetwork(TestCase):
    """Test the Neural Netork Policy."""

    fields = ['args', 'kwargs', 'action_space', 'state_space', 'dtype',
              'layers', 'scope', 'init_weights', 'activation', 'X', 'a',
              'W_action', 'W_var', 'a_pred', 'var', 'h', 'is_set_up']

    def test_initialization(self):
        """Test: NEURALNETWORK: initialization."""
        # test bad layer size:
        args = [[2]]
        with self.assertRaises(ValueError):
            NeuralNetwork(*args)

        # test field existence
        args = [[2, 6, 1]]

        nn = NeuralNetwork(*args)

        for field in self.fields:
            assert hasattr(nn, field)

        # test network setup
        kwargs = {
            'do_setup': True
        }

        nn = NeuralNetwork(*args, **kwargs)

        # check field contents.
        assert(all([a == b for a, b in zip(args, nn.args)]))
        self.assertEqual(nn.layers, args[0])
        self.assertEqual(nn.dtype, 'float')

        self.assertEqual(len(nn.W_action), 2)
        self.assertEqual(len(nn.W_var), 1)

        # well... because is does not work for whatever fucking reason.
        self.assertEqual(str(type(nn.a_pred)), str(tf.Tensor))
        self.assertIn(str(type(nn.var)), (str(tf.Tensor), str(tf.constant)))

        self.assertEqual(len(nn.h), 2)

    def test_mapping(self):
        """Test: NEURALNETWORK: mapping."""
        args = [[2, 1]]

        kwargs = {
            'weights': [tf.constant([2., 1.], shape=(2, 1))],
            'do_setup': True,
        }

        nn = NeuralNetwork(*args, **kwargs)

        sess = tf.Session()

        with sess.as_default():
            self.assertEqual(nn(np.array([2., 1.])), [5.])

    def test_variable_assignment(self):
        """Test: NEURALNETWORK: parameter assignment."""
        args = [[2, 1]]
        kwargs = {'do_setup': True}

        nn = NeuralNetwork(*args, **kwargs)

        with tf.Session().as_default():
            nn.parameters = nn.W_action[0].assign([[2.], [1.]])
            assert((np.array([[2.], [1.]]) == nn.parameters).all())
            self.assertEqual(nn(np.array([2., 1.])), [5.])

    def test_copy(self):
        """Test: NEURALNETWORK: copy."""
        nn = NeuralNetwork([2, 6, 1])
        nn_copy = nn.copy(scope='copy', do_setup=False)

        exclude = ('scope', 'kwargs')

        for field in self.fields:
            if field not in exclude and field in nn.kwargs.keys():
                print(field)
                self.assertEquals(getattr(nn, field, None),
                                  getattr(nn_copy, field, None))


class TestLinearPolicy(TestCase):
    """Test the Linear Policy."""

    def test_initialization(self):
        """Test: LINEARPOLICY: initialization."""
        lp = LinearPolicy(2, 1)

        self.assertEqual(lp.d_state, 2)
        self.assertEqual(lp.d_action, 1)

        self.assertEqual(lp.par_dim, 2)
        self.assertIs(lp._par_space, None)

        self.assertFalse(lp.initialized)

        self.assertIs(lp._parameters, None)
        self.assertTrue(lp.biased)
        self.assertEqual(lp._bias, 0)
        self.assertIs(lp._par, None)

        par_mock = Mock()
        par_space_mock = Mock()

        with self.assertRaises(ValueError):
            lp_mocked = LinearPolicy(2, 1, par_mock, par_space_mock)

        par_mock = [2, 1]

        lp_mocked = LinearPolicy(2, 1, par_mock, par_space_mock)

        self.assertTrue(lp_mocked.initialized)
        assert(all(par_mock == lp_mocked.parameters))

        self.assertEqual(par_space_mock, lp_mocked.parameter_space)

    def test_discrete_map(self):
        """Test: DISCRETELINEARPOLICY: map."""
        dp = DiscreteLinearPolicy(2, 1, biased=False)
        dp.parameters = np.array([1, 1])
        self.assertEqual(dp([1, 1]), 1)
        self.assertEqual(dp([-1, -1]), 0)

        dp2 = DiscreteLinearPolicy(2, 2, biased=False)
        dp2.parameters = np.array([1, 1, -1, -1])
        assert(all(dp2([1, 1]) == [1, 0]))
        assert(all(dp2([-1, -1]) == [0, 1]))


class TestController(TestCase):
    """Test NonLinearQuadrocopterController."""

    def test_controller_init(self):
        """Test: CONTROLLER: initialization."""
        ctrl = NonLinearQuadrocopterController()

        self.assertEquals(ctrl._zeta_z, .7)
        assert(all(isclose(ctrl._params, [.7, .7, .7, .5, .707])))
        self.assertIsNone(ctrl.reference)
        self.assertTrue(ctrl.initialized)
        self.assertIsInstance(ctrl._par_space, BoundedSpace)

    def test_controller_map(self):
        """Test: CONTROLLER: mapping."""
        ref = Reference('circle', 1 / 70.)
        ref.reset(StateVector())
        ctrl = NonLinearQuadrocopterController(reference=ref)

        action = ctrl(StateVector())

        print(action)
        assert all(isclose(action, [0.20510876, -0.30667618, 0., -6.28318548]))

    def test_controller_properties(self):
        """Test: CONTROLLER: properties."""
        ctrl = NonLinearQuadrocopterController()

        ctrl.parameters = [0., 1., 0., 1., 0.]
        assert(all(np.isclose(ctrl.parameters, [0., 1., 0., 1., 0.])))

        self.assertEquals(ctrl.parameter_space, ctrl._par_space)



from .linear_policy import LinearPolicy, NoisyLinearPolicy
from .linear_policy import DiscreteLinearPolicy
from .neural_network import NeuralNetwork
from .controller import NonLinearQuadrocopterController

__all__ = [
    'LinearPolicy',
    'NoisyLinearPolicy',
    'DiscreteLinearPolicy',
    'NeuralNetwork',
    'NonLinearQuadrocopterController'
]



"""SafeOpt Wrapper."""

from SafeRLBench import AlgorithmBase
from SafeRLBench.error import add_dependency

from numpy import mean, array

try:
    import safeopt
except:
    safeopt = None

try:
    import GPy
except:
    GPy = None

import logging

logger = logging.getLogger(__name__)

__all__ = ('SafeOpt', 'SafeOptSwarm')


class _SafeOptWrap(AlgorithmBase):

    def __init__(self, opt, gp_opt_par, gp_par, environment, policy, max_it,
                 avg_reward, window):
        super(_SafeOptWrap, self).__init__(environment, policy, max_it)

        self._opt = opt

        self.gp_opt = None

        self.gp_opt_par = gp_opt_par
        self.gp_par = gp_par

        self.avg_reward = avg_reward
        self.window = window
        self.rewards = []

    def _initialize(self):
        logger.debug("Initializing Policy.")
        # check if policy is already initialized by the user
        if self.policy.initialized:
            logger.debug("Use pre-set policy parameters.")
            parameters = self.policy.parameters
        else:
            logger.debug("Draw parameters at random.")
            parameters = self.policy.parameter_space.sample()
            self.policy.parameters = parameters

        # Compute a rollout
        trace = self.environment.rollout(self.policy)
        reward = sum([t[2] for t in trace])

        # Initialize gaussian process with args:
        gp = []
        for pars in zip(*self.gp_par):
            gp.append(GPy.core.GP(array([parameters]), array([[reward]]),
                      *pars))

        # Initialize SafeOpt
        self.gp_opt = self._opt(gp, **self.gp_opt_par)

    def _step(self):
        parameters = self.gp_opt.optimize()
        self.policy.parameters = parameters

        trace = self.environment.rollout(self.policy)
        reward = sum([t[2] for t in trace])

        self.gp_opt.add_new_data_point(parameters, reward)
        self.rewards.append(reward)

    def _is_finished(self):
        if ((len(self.rewards) > self.window)
                and mean(self.rewards[(len(self.rewards) - self.window):-1])
                > self.avg_reward):
            return True
        else:
            return False


class SafeOpt(_SafeOptWrap):
    """Wrap SafeOpt algorithm.

    This class wraps the `SafeOpt` algorithm. It relies on the original
    implementation of `SafeOpt` which has to be installed before using this
    wrapper.

    Attributes
    ----------
    environment :
        Environment to be optimized.
    policy :
        Policy to be optimized.
    max_it :
        Maximal number of iterations before we abort.
    avg_reward : integer
        Average reward at which the optimization will be finished.
    window : integer
        Window for the average reward
    gp : GPy Gaussian process
        A Gaussian process which is initialized with safe, initial data points.
        If a list of GPs then the first one is the value, while all the
        other ones are safety constraints.
    gp_opt : SafeOptSwarm object
        Instance of `SafeOptSwarm` used for optimization.
    gp_opt_par : dict
        Dictionary of parameters to initialize `SafeOpt`.
    """

    def __init__(self,
                 environment, policy, max_it, avg_reward, window,
                 kernel, likelihood, parameter_set, fmin,
                 lipschitz=None, beta=3.0, num_contexts=0, threshold=0,
                 scaling='auto', info=None):
        """Initialize Attributes.

        Parameters
        ----------
        environment :
            environmet to be optimized.
        policy :
            policy to be optimized.
        max_it :
            maximal number of iterations before we abort.
        avg_reward : integer
            average reward at which the optimization will be finished.
        window : integer
            window for the average reward
        kernel : GPy kernel
            Kernel used to initialize the gaussian process. If this is a list
            multiple kernels will be initialized. The size of this argument
            has to agree with the size of the likelihood.
        likelihood : GPy likelihood
            Likelihood used to initialize kernels. If this is a list, multiple
            kernels will be initialized. The size of this argument has to
            agree with the size of the likelihood.
        parameter_set : 2d-array
            List of parameters
        fmin : list of floats
            Safety threshold for the function value. If multiple safety
            constraints are used this can also be a list of floats (the first
            one is always the one for the values, can be set to None if not
            wanted)
        lipschitz : list of floats
            The Lipschitz constant of the system, if None the GP confidence
            intervals are used directly.
        beta : float or callable
            A constant or a function of the time step that scales the
            confidence interval of the acquisition function.
        threshold : float or list of floats
            The algorithm will not try to expand any points that are below this
            threshold. This makes the algorithm stop expanding points
            eventually. If a list, this represents the stopping criterion for
            all the gps. This ignores the scaling factor.
        scaling : list of floats or "auto"
            A list used to scale the GP uncertainties to compensate for
            different input sizes. This should be set to the maximal variance
            of each kernel. You should probably leave this to "auto" unless
            your kernel is non-stationary.
        info :
            Dummy argument that can hold anything usable to identify the
            configuration.
        """
        add_dependency(safeopt, 'SafeOpt')
        add_dependency(GPy, 'GPy')

        # store the `SafeOpt` arguments.
        gp_opt_par = {
            'parameter_set': parameter_set,
            'fmin': fmin,
            'lipschitz': lipschitz,
            'beta': beta,
            'num_contexts': num_contexts,
            'threshold': threshold,
            'scaling': scaling}

        # store the kernel arguments
        if not isinstance(kernel, list):
            kernel = [kernel]
        if not isinstance(likelihood, list):
            likelihood = [likelihood]
        assert len(likelihood) == len(kernel), (
            'kernel and likelihood need to have same length (%d /= %d)'
            % (len(likelihood), len(kernel)))

        gp_par = (kernel, likelihood)

        super(SafeOpt, self).__init__(safeopt.SafeOpt, gp_opt_par, gp_par,
                                      environment, policy, max_it, avg_reward,
                                      window)


class SafeOptSwarm(_SafeOptWrap):
    """Wrap SafeOpt algorithm.

    This class wraps the `SafeOptSwarm` algorithm. It relies on the original
    implementation of `SafeOptSwarm` which is part of the `safeopt` package
    and has to be installed before using this class.

    Attributes
    ----------
    environment :
        Environment to be optimized.
    policy :
        Policy to be optimized.
    max_it :
        Maximal number of iterations before we abort.
    avg_reward : integer
        Average reward at which the optimization will be finished.
    window : integer
        Window for the average reward
    gp : GPy Gaussian process
        A Gaussian process which is initialized with safe, initial data points.
        If a list of GPs then the first one is the value, while all the
        other ones are safety constraints.
    gp_opt : SafeOptSwarm object
        Instance of SafeOptSwarm used for optimization.
    gp_opt_par : list
        List of parameters to initialize `SafeOpt`.
    """

    def __init__(self,
                 environment, policy, max_it, avg_reward, window,
                 kernel, likelihood, fmin, bounds, beta=3.0, threshold=0,
                 scaling='auto', swarm_size=20, info=None):
        """Initialize Attributes.

        Parameters
        ----------
        environment :
            Environment to be optimized.
        policy :
            policy to be optimized.
        max_it :
            maximal number of iterations before we abort.
        avg_reward : integer
            average reward at which the optimization will be finished.
        window : integer
            window for the average reward
        kernel : GPy kernel
            Kernel used to initialize the gaussian process. If this is a list
            multiple kernels will be initialized. The size of this argument
            has to agree with the size of the likelihood.
        likelihood : GPy likelihood
            Likelihood used to initialize kernels. If this is a list, multiple
            kernels will be initialized. The size of this argument has to
            agree with the size of the likelihood.
        fmin : list of floats
            Safety threshold for the function value. If multiple safety
            constraints are used this can also be a list of floats (the first
            one is always the one for the values, can be set to None if not
            wanted)
        bounds : pair of floats or list of pairs of floats
            If a list is given, then each pair represents the lower/upper bound
            in each dimension. Otherwise, we assume the same bounds for all
            dimensions. This is mostly important for plotting or to restrict
            particles to a certain domain.
        beta : float or callable
            A constant or a function of the time step that scales the
            confidence interval of the acquisition function.
        threshold : float or list of floats
            The algorithm will not try to expand any points that are below this
            threshold. This makes the algorithm stop expanding points
            eventually. If a list, this represents the stopping criterion for
            all the gps. This ignores the scaling factor.
        scaling : list of floats or "auto"
            A list used to scale the GP uncertainties to compensate for
            different input sizes. This should be set to the maximal variance
            of each kernel. You should probably set this to "auto" unless your
            kernel is non-stationary
        swarm_size : int
            The number of particles in each of the optimization swarms
        info :
            Dummy argument that can hold anything usable to identify the
            configuration.
        """
        add_dependency(safeopt, 'SafeOpt')
        add_dependency(GPy, 'GPy')

        # store the `SafeOpt` arguments.
        gp_opt_par = {
            'fmin': fmin,
            'bounds': bounds,
            'beta': beta,
            'threshold': threshold,
            'scaling': scaling,
            'swarm_size': swarm_size
        }

        # store the kernel arguments
        if not isinstance(kernel, list):
            kernel = [kernel]
        if not isinstance(likelihood, list):
            likelihood = [likelihood]
        assert len(likelihood) == len(kernel), (
            'kernel and likelihood need to have same length (%d /= %d)'
            % (len(likelihood), len(kernel)))

        gp_par = (kernel, likelihood)

        super(SafeOptSwarm, self).__init__(safeopt.SafeOptSwarm, gp_opt_par,
                                           gp_par, environment, policy, max_it,
                                           avg_reward, window)



"""Markov Decision Process Implementations."""

import numpy as np

from SafeRLBench import EnvironmentBase
from SafeRLBench.spaces import DiscreteSpace


class MDP(EnvironmentBase):
    """Discrete Markov Decision Process Environment.

    Attributes
    ----------
    transitions : array-like
        Array holding transition matrix for each action. The dimension of
        the state and action spaces will be deduced from this array.
    rewards : array-like
        Array holding the reward matrix for each action. It needs to comply
        with the dimensions deduced from the transitions array.
    action_space : DiscreteSpace object
        Action space as determined from the transitions array
    state_space : DiscreteSpace object
        State space as determined from the transitions array.
    init_state : int
        Initial state of the process. If None, it will be set to 0.
    state : int
        Current state of the system.
    """

    def __init__(self, transitions, rewards, horizon=100, init_state=None,
                 seed=None):
        """MDP initialization.

        Parameters
        ----------
        transitions : array-like
            Array holding transition matrix for each action. The dimension of
            the state and action spaces will be deduced from this array.
        rewards : array-like
            Array holding the reward matrix for each action. It needs to comply
            with the dimensions deduced from the transitions array.
        init_state : int
            Initial state of the process. If None, it will be set to 0.
        """
        self.horizon = horizon

        self.transitions = transitions
        self.rewards = rewards

        # determine state and action space
        self.action_space = DiscreteSpace(len(transitions))
        self.state_space = DiscreteSpace(len(transitions[0]))

        # if initial state is none, we will use 0 as an initial state
        if init_state is None:
            init_state = 0
        elif not self.state_space.contains(init_state):
            raise ValueError('Initial state (%d) is not a valid state.',
                             init_state)

        # setup current state and store the initial state for reset
        self.init_state = init_state
        self.state = init_state

        # generate random state
        self.random = np.random.RandomState()

        if seed is not None:
            self.seed = seed
        else:
            self._seed = None

    @property
    def seed(self):
        """Seed."""
        return self._seed

    @seed.setter
    def seed(self, v):
        self.random.seed(v)
        self._seed = v

    def _update(self, action):
        prev_state = self.state

        # choose next state
        self.state = self.random.choice(np.arange(self.state_space.dimension),
                                        p=self.transitions[action][self.state])
        # determine reward
        reward = self.rewards[action][prev_state][self.state]

        return action, self.state, reward

    def _reset(self):
        self.state = self.init_state


def _get_test_args():
    # private method that will generate arguments for mdp testing.
    transitions = [
        [[.1, .9, 0., 0., 0.],
         [.2, 0., .8, 0., 0.],
         [.3, 0., 0., .7, 0.],
         [.4, 0., 0., 0., .6],
         [.4, 0., 0., 0., .6]],
        [[1., 0., 0., 0., 0.],
         [1., 0., 0., 0., 0.],
         [1., 0., 0., 0., 0.],
         [1., 0., 0., 0., 0.],
         [1., 0., 0., 0., 0.]]
    ]

    rewards = [
        [[0., 0., 0., 0., 0.],
         [0., 0., 0., 0., 0.],
         [0., 0., 0., 0., 0.],
         [0., 0., 0., 0., 0.],
         [0., 0., 0., 0., 0.]],
        [[0., 0., 0., 0., 0.],
         [1., 0., 0., 0., 0.],
         [2., 0., 0., 0., 0.],
         [3., 0., 0., 0., 0.],
         [4., 0., 0., 0., 0.]],
    ]

    return [transitions, rewards, 100, None, 42]



"""Linear Policy Class."""

from SafeRLBench import Policy, ProbPolicy
from SafeRLBench.spaces import BoundedSpace

import numpy as np

__all__ = ('LinearPolicy', 'DiscreteLinearPolicy', 'NoisyLinearPolicy')


class LinearPolicy(Policy):
    """Policy implementing a linear mapping from state to action space.

    Attributes
    ----------
    d_state : positive integer
        Dimension of the state space.
    d_action : positive integer
        Dimension of the action space
    parameters : nd-array
        Array containing initial parameters.
    initialized : boolean
        Boolean indicating if parameters have been initialized.
    biased : boolean
        Flag indicating if the policy is supposed to be biased or not.
    """

    def __init__(self, d_state, d_action,
                 par=None, par_space=None, biased=True):
        """Initialize LinearPolicy.

        Parameters
        ----------
        d_state : positive integer
            Dimension of the state space.
        d_action : positive integer
            Dimension of the action space
        par : ndarray
            Array containing initial parameters. If there is a constant bias,
            the array needs to be flat with shape (d_state * d_action + 1,).
            Otherwise it may either have shape (d_action, d_state) or
            (d_state * d_action,)
        biased : boolean
            Flag indicating if the policy is supposed to be biased or not.
        """
        assert(d_state > 0 and d_action > 0)
        self.d_state = d_state
        self.d_action = d_action

        self.par_dim = d_state * d_action

        self._par_space = None

        self.initialized = False

        if par is not None:
            self.parameters = par
        else:
            # make sure some fields exist.
            self._parameters = None
            self.biased = biased
            self._bias = 0
            self._par = None

        if par_space is not None:
            self.parameter_space = par_space

    def map(self, state):
        """Map a state to an action.

        Parameters
        ----------
        state : array-like
            Element of state space.

        Returns
        -------
        action : ndarray
            Element of action space.
        """
        if self.d_action == 1:
            ret = self._parameters.dot(state).item() + self._bias
        else:
            ret = self._parameters.dot(state) + self._bias
        return ret

    @property
    def parameters(self):
        """Property to access parameters.

        The property returns the same representation as used when set.
        If the mapping contains a bias, then the input needs to be a ndarray
        with shape (d_action * d_state + 1,) otherwise it may either be a
        (d_action, d_state) or (d_action * d_state,) shaped array
        """
        if not self.initialized:
            raise NameError('Policy parameters not initialized yet.')
        return self._par

    @parameters.setter
    def parameters(self, par):
        par = np.array(par).copy()

        if not self.initialized:
            shape = par.shape
            if (shape == (self.d_action, self.d_state)
                    or shape == (self.par_dim,)):
                self.biased = False
                self._bias = 0
            elif shape == (self.par_dim + 1,):
                self.biased = True
            else:
                raise ValueError("Parameters with shape %s invalid.",
                                 str(shape))

            self.initialized = True

        # store parameter in original representation.
        self._par = par

        if self.d_action == 1:
            shape = (self.d_state,)
        else:
            shape = (self.d_action, self.d_state)

        if not self.biased:
            self._parameters = par.reshape(shape)
        else:
            self._bias = par[-1]
            self._parameters = par[0:-1].reshape(shape)

    @property
    def parameter_space(self):
        """Property storing the parameter space.

        By default the parameter space will be assigned to be a BoundedSpace
        between [0,1]^d. However it might be necessary to change this. A user
        may thus assign a new parameter space.

        WARNING: Currently there is no sanity check for manually assigned
        parameter spaces.
        """
        if self._par_space is None:
            if self.biased:
                shape = (self.par_dim + 1,)
            else:
                shape = (self.par_dim,)
            self._par_space = BoundedSpace(0, 1, shape)

        return self._par_space

    @parameter_space.setter
    def parameter_space(self, par_space):
        self._par_space = par_space


class DiscreteLinearPolicy(LinearPolicy):
    """LinearPolicy on a discrete action space of {0, 1}^d."""

    def map(self, state):
        """Map to discrete action space.

        Parameters
        ----------
        state : element of state space
            state to be mapped.

        Returns
        -------
        action : ndarray
            Element of {0, 1}^d_action
        """
        cont_action = super(DiscreteLinearPolicy, self).map(state)
        if self.d_action == 1:
            if (cont_action < 0):
                action = 0
            else:
                action = 1
        else:
            action = np.zeros(cont_action.shape, dtype=int)
            action[cont_action > 0] += 1

        return action


class NoisyLinearPolicy(LinearPolicy, ProbPolicy):
    """
    Policy implementing a linear mapping from state to action space with noise.

    Attributes
    ----------
    d_state : positive integer
        Dimension of the state space.
    d_action : positive integer
        Dimension of the action space
    sigma : double
        Sigma for gaussian noise
    parameters : nd-array
        Array containing initial parameters.
    initialized : boolean
        Boolean indicating if parameters have been initialized.
    biased : boolean
        Flag indicating if the policy is supposed to be biased or not.
    """

    def __init__(self, d_state, d_action, sigma,
                 par=None, par_space=None, biased=False):
        """Initialize Noisy Linear Policy.

        Parameters
        ----------
        d_state : positive integer
            Dimension of the state space.
        d_action : positive integer
            Dimension of the action space
        sigma : double
            Sigma for gaussian noise
        par : ndarray
            Array containing initial parameters. If there is a constant bias,
            the array needs to be flat with shape (d_state * d_action + 1,).
            Otherwise it may either have shape (d_action, d_state) or
            (d_state * d_action,)
        biased : boolean
            Flag indicating if the policy is supposed to be biased or not.
        """
        assert(d_state > 0 and d_action > 0)

        self.sigma = sigma

        self.random_state = np.random.RandomState()

        super(NoisyLinearPolicy, self).__init__(d_state, d_action, par,
                                                par_space, biased)

    def map(self, state):
        """Map a state to an action.

        Parameters
        ----------
        state : array-like
            Element of state space.

        Returns
        -------
        action : ndarray
            Element of action space.
        """
        noise = self.random_state.normal(0, self.sigma)
        return super(NoisyLinearPolicy, self).map(state) + noise

    def grad_log_prob(self, state, action):
        """Compute the gradient of the logarithm of the probability dist."""
        noise = action - super(NoisyLinearPolicy, self).map(state)
        return - 2 * noise * self.parameters / self.sigma**2



"""Asynchronous Actor-Critic Agents.

Implementations refer to Denny Britz implementations at
https://github.com/dennybritz/reinforcement-learning/tree/master/PolicyGradient/a3c
"""

import copy
import threading

import numpy as np

from SafeRLBench import AlgorithmBase

from SafeRLBench.error import add_dependency

try:
    import tensorflow as tf
    from tensorflow.contrib.distributions import Normal
except:
    tf = None

import logging

logger = logging.getLogger(__name__)


def _run_thread(alg, worker, sess, coord):
    while not coord.should_stop() and not alg.is_finished():
        p_net_loss, v_net_loss = worker.run(sess)
        alg.step()

    coord.request_stop()


class A3C(AlgorithmBase):
    """Implementation of the Asynchronous Actor-Critic Agents Algorithm.

    Here we use several workers threads, which asynchronously update their
    local estimation networks. The networks consist of a policy estimation
    network which estimates the policy and a value estimation network which
    is used to estimate the expected reward.

    Attributes
    ----------
    environment :
        Environment we want to optimize the policy on.
    policy :
        The policy we want to optimize. The policy has to comply with some
        special requirements to run with this algorithm. It needs to
        implement a tensorflow graph with placeholders stored in the
        attributes X, for the input, and a, for the output.
        Further it will need to implement a ``setup()`` method which will
        effectively assemble the neural network, as well as a ``copy()``
        method that will generate a copy of the network for the worker
        threads. Unfortunately there is no base class for this yet, see
        the ``NeuralNetwork`` policy as a reference implementation.
    max_it : integer
        The maximal number of iterations before we abort.
    num_workers : integer
        Number of workers that should be used asynchronous.
    rate : float
        Rate for Gradient Descent.
    discount : float
        Discount factor for adjusted reward.
    log_file : string
        Indicate the relative path to directory where a summary event file
        should be generated. If `None` no tensorflow logs will be stored.
    done : boolean
        Indicates whether the run is done.
    workers : list
        List containing the worker instances.
    threads : list
        List containing the thread instances

    Notes
    -----
    A proper neural network policy base class does not exist yet, but would
    be very nice to have.

    Bug:
        A3C does not properly run parallel. Actually according to tensorflow
        the GIL should be release upon run invocation, but every thread but
        one just stalls until one worker finished everything on his own.
    """

    def __init__(self, environment, policy, max_it=1000, num_workers=1,
                 rate=0.1, discount=0.1, log_file=False):
        """Initialize A3C.

        Parameters
        ----------
        environment:
            Environment we want to optimize the policy on.
        policy:
            The policy we want to optimize. The policy needs be defined by a
            tensorflow neural network and define certain attributes.
        max_it: int
            Maximum number of iterations.
        num_workers: int
            Number of workers.
        rate: float
            Update rate passed to the optimizer.
        discount: float
            Discount for the computation of the discounted reward.
        log_file: string
            Indicate the relative path to directory where a summary event file
            should be generated. If `None` no tensorflow logs will be stored.
        """
        add_dependency(tf, 'TensorFlow')

        if policy.is_set_up:
            raise(ValueError('Policy should not be set up.'))

        super(A3C, self).__init__(environment, policy, max_it)

        self.num_workers = num_workers
        self.rate = rate
        self.discount = discount

        self.done = False

        self.log_file = log_file

        self.policy = policy

        # init networks
        with tf.device("/cpu:0"):
            with tf.variable_scope('global'):
                self.p_net = _PolicyNet(self.policy, rate)
                self.v_net = _ValueNet(self.policy, rate)

        self.workers = []
        self.threads = []

        self.global_counter = 0

        self.sess = None
        self.coord = None

    def _initialize(self):
        self.global_counter = 0
        self.workers = []
        self.threads = []

        self.done = False

        self.coord = tf.train.Coordinator()

        for i in range(self.num_workers):
            worker = _Worker(self.environment,
                             self.policy,
                             self.p_net,
                             self.v_net,
                             self.discount,
                             'worker_' + str(i))
            self.workers.append(worker)

        init_op = tf.global_variables_initializer()
        self.sess = tf.Session()
        self.sess.run(init_op)

        self.policy.sess = self.sess

        # Write a graph file
        if self.log_file:
            graph = self.sess.graph
            writer = tf.summary.FileWriter(self.log_file, graph=graph)
            writer.flush()

    def _step(self):
        self.global_counter += 1

        if self.global_counter % 10 == 0:
            logger.debug("Global update counter at step %d.",
                         self.global_counter)

    def _is_finished(self):
        if self.global_counter >= self.max_it:
            self.done = True
        return self.done

    def _optimize(self):
        self.initialize()
        for worker in self.workers:
            t = threading.Thread(target=_run_thread(self, worker, self.sess,
                                                    self.coord))
            self.threads.append(t)

        for t in self.threads:
            t.start()
        for t in self.threads:
            t.join()


class _Worker(object):
    """Worker thread."""

    def __init__(self, env, policy, p_net, v_net, discount, name):
        self.name = name
        self.env = copy.copy(env)
        self.global_policy = policy
        self.global_p_net = p_net
        self.global_v_net = v_net

        self.discount = discount

        # generate local networks
        self.local_policy = policy.copy(name, do_setup=False)
        with tf.variable_scope(name):
            self.local_p_net = _PolicyNet(self.local_policy,
                                          self.global_p_net.rate)
            self.local_v_net = _ValueNet(self.local_policy,
                                         self.global_v_net.rate)

        # create copy op
        trainable_variables = tf.GraphKeys.TRAINABLE_VARIABLES
        self.copy_params_op = self.make_copy_params_op(
            tf.contrib.slim.get_variables(scope="global",
                                          collection=trainable_variables),
            tf.contrib.slim.get_variables(scope=self.name,
                                          collection=trainable_variables))

        # create train ops
        self.p_net_train = self.make_train_op(self.local_p_net,
                                              self.global_p_net)
        self.v_net_train = self.make_train_op(self.local_v_net,
                                              self.global_v_net)

        self.state = self.env.state

    def run(self, sess):
        with sess.as_default():
            sess.run(self.copy_params_op)

            # perform a rollout
            trace = self.env.rollout(self.local_policy)

            advantages = []
            values = []
            states = []
            actions = []

            value = 0.

            tot_reward = 0.
            for (action, state, reward) in trace:
                state = state.flatten()

                value = reward + self.discount * value

                # evaluate value net on state
                value_pred = sess.run(self.local_v_net.V_est,
                                      {self.local_v_net.X: [state]})
                advantage = reward - value_pred.flatten()

                tot_reward += reward

                advantages.append(advantage)
                values.append(value)
                states.append(state)
                actions.append(action)

            print(tot_reward)

            # compute local gradients and train global network
            feed_dict = {
                self.local_p_net.X: np.array(states),
                self.local_p_net.target: advantages,
                self.local_p_net.a: actions,
                self.local_v_net.X: np.array(states),
                self.local_v_net.V: values
            }

            p_net_loss, v_net_loss, _, _ = sess.run([
                self.local_p_net.loss,
                self.local_v_net.loss,
                self.p_net_train,
                self.v_net_train
            ], feed_dict)

        return p_net_loss, v_net_loss

    @staticmethod
    def make_copy_params_op(v1_list, v2_list):
        """Create operation to copy parameters.

        Creates an operation that copies parameters from variable in v1_list to
        variables in v2_list.
        The ordering of the variables in the lists must be identical.
        """
        v1_list = list(sorted(v1_list, key=lambda v: v.name))
        v2_list = list(sorted(v2_list, key=lambda v: v.name))

        update_ops = []
        for v1, v2 in zip(v1_list, v2_list):
            op = v2.assign(v1)
            update_ops.append(op)

        return update_ops

    @staticmethod
    def make_train_op(loc, glob):
        """Create operation that applies local gradients to global network."""
        loc_grads, _ = zip(*loc.grads_and_vars)
        loc_grads, _ = tf.clip_by_global_norm(loc_grads, 5.0)
        _, glob_vars = zip(*glob.grads_and_vars)
        loc_grads_glob_vars = list(zip(loc_grads, glob_vars))
        get_global_step = tf.contrib.framework.get_global_step()

        return glob.opt.apply_gradients(loc_grads_glob_vars,
                                        global_step=get_global_step)


class _ValueNet(object):
    """Wrapper for the Value function."""

    def __init__(self, policy, rate, train=True):
        self.rate = rate

        with tf.variable_scope('value_estimator'):
            self.X = tf.placeholder(policy.dtype,
                                    shape=policy.X.shape,
                                    name='X')
            self.V = tf.placeholder(policy.dtype,
                                    shape=[None],
                                    name='V')

            self.W = policy.init_weights((policy.layers[0], 1))

            self.V_est = tf.matmul(self.X, self.W)

            self.losses = tf.squared_difference(self.V_est, self.V)
            self.loss = tf.reduce_sum(self.losses, name='loss')

            if train:
                self.opt = tf.train.RMSPropOptimizer(rate, 0.99, 0.0, 1e-6)
                self.grads_and_vars = self.opt.compute_gradients(self.loss)
                self.grads_and_vars = [(g, v) for g, v in self.grads_and_vars
                                       if g is not None]
                self.update = self.opt.apply_gradients(self.grads_and_vars)


class _PolicyNet(object):
    """Wrapper for the Policy function."""

    def __init__(self, policy, rate, train=True):
        self.rate = rate
        self.policy = policy

        with tf.variable_scope('policy_estimator'):
            self.policy.setup()

            self.X = policy.X
            self.a = policy.a
            self.target = tf.placeholder(dtype='float', shape=[None, 1],
                                         name='target')

            self.a_pred = policy.a_pred
            self.var = policy.var

            dist = Normal(self.a_pred, self.var)
            self.log_probs = dist.log_pdf(self.a)

            self.losses = self.log_probs * self.target
            self.loss = tf.reduce_sum(self.losses, name='loss')

            if train:
                self.opt = tf.train.RMSPropOptimizer(rate, 0.99, 0.0, 1e-6)
                self.grads_and_vars = self.opt.compute_gradients(self.loss)
                self.grads_and_vars = [(g, v) for g, v in self.grads_and_vars
                                       if g is not None]
                self.update = self.opt.apply_gradients(self.grads_and_vars)



from __future__ import division, print_function, absolute_import

from .rd_space import RdSpace
from .bounded_space import BoundedSpace
from .discrete_space import DiscreteSpace

__all__ = ['RdSpace', 'BoundedSpace', 'DiscreteSpace']



from setuptools import setup

setup(
    name='SafeRLBench',
    version='1.0.1',
    author='Nicolas Ochsner',
    author_email='ochsnern@student.ethz.ch',
    packages=[
        'SafeRLBench',
        'SafeRLBench.algo',
        'SafeRLBench.envs',
        'SafeRLBench.spaces',
        'SafeRLBench.policy',
    ],
    description='Safe Reinforcement Learning Benchmark',
    keywords='reinforcement-learning benchmark',
    url='https://github.com/befelix/Safe-RL-Benchmark',
    install_requires=[
        'numpy >= 1.7',
        'scipy >= 0.19.0',
        'six >= 1.10',
        'futures >= 3.0.5;python_version<"3.2"'
    ],
    extras_require={
        'gym': ['gym >= 0.8.0'],
        'safeopt': ['GPy >= 1.6.1', 'safeopt >= 0.1'],
        'neural': ['tensorflow >= 1.0.0'],
    },
    dependency_links=[
        'git+https://github.com/befelix/SafeOpt/tarball/master#egg=safeopt-0.1'
    ],
)



"""Wrapper for OpenAI Gym."""

from SafeRLBench import EnvironmentBase
from SafeRLBench.error import add_dependency

try:
    import gym
except:
    gym = None


# TODO: GymWrap: Add examples to docs
class GymWrap(EnvironmentBase):
    """Wrapper class for the OpenAI Gym.

    Attributes
    ----------
    env : gym environment
        Environment of the OpenAI Gym created by gym.make().
    horizon : integer
        Horizon for rollout.
    render : boolean
        Default: False. If True simulation will be rendered during rollouts on
        this instance.

    Notes
    -----
    The GymWrap class relies on the complete observability of the state
    through a state field in the respective gym environment. For the classic
    control problem this is indeed the case, but on other environment it
    remains to be untested.
    """

    def __init__(self, env, horizon=100, render=False):
        """Initialize attributes.

        Parameters
        ----------
        env : gym environment
            Instance of the gym environment that should be optimized on.
        horizon : integer
            Horizon for rollout.
        render : boolean
            Default: False ; If True simulation will be rendered during
            rollouts on this instance.
        """
        add_dependency(gym, 'Gym')

        EnvironmentBase.__init__(self, env.observation_space, env.action_space,
                                 horizon)
        self.environment = env.unwrapped
        self.render = render
        self.done = False

        self.environment.reset()

    def _update(self, action):
        observation, reward, done, info = self.environment.step(action)
        self.done = done
        return action, observation, reward

    def _reset(self):
        self.environment.reset()
        self.done = False

    def _rollout(self, policy):
        trace = []
        for n in range(self.horizon):
            if self.render:
                self.environment.render()
            trace.append(self.update(policy(self.state)))
            if self.done:
                break
        return trace

    @property
    def state(self):
        """Observable system state."""
        return self.environment.state

    @state.setter
    def state(self, s):
        assert self.state_space.contains(s)
        self.environment.state = s


def _get_test_args():
    return [gym.make('MountainCar-v0')]



"""Neural Network Policy implementation."""

from SafeRLBench import Policy
from SafeRLBench.error import add_dependency, MultipleCallsException
from SafeRLBench.spaces import RdSpace

import numpy as np
from numpy.random import normal

try:
    import tensorflow as tf
except:
    tf = None

import logging

logger = logging.getLogger(__name__)


def default_init_weights(shape):
    """Default weights initialization."""
    weights = tf.random_normal(shape, mean=0, stddev=0.1, name='weights')
    return tf.Variable(weights)


class NeuralNetwork(Policy):
    """Fully connected Neural Network Policy.

    Attributes
    ----------
    args : list
        Contains the args used to initialize the policy.
    kwargs : dict
        Contains the kwargs used to initialize the policy.
    layers : list of integers
        A list describing the layer sizes. The first element represents the
        size of the input layer, the last element the size of the output
        layer.
    state_space : space instance
    action_space : space instance
    weights : tf.Variable
        If none the init_weights function will be used to initialize the
        weights.
    init_weights : callable
        Takes a shape as an argument and returns a tf.Variable according to
        this shape.
    activation : list of activation functions
        An activation function which will be used to construct the respective
        layer. If only one activation function is passed, it will be used for
        every layer. If the argument is None by default the sigmoid function
        will be used.
    dtype : string
        Data type of input and output.
    W_action : list of tf.Variable
        The list contains the `tf.Variable` instances describing the mapping
        between the hidden layers. The i-th entry describes the connection
        between layer i and layer i+1.
    W_var : list of tf.Variable
        This list contains the weights used to compute the variance estimation.
        Each entry corresponds to one layer and contains weights of shape
        (layer[i], 1).
    a_pred :
        Action estimate of the fully connected neural network defined by
        `W_action` and activation.
    var :
        Variance estimate which is a weighted sum of all hidden units.
        The weights are described by `W_var`.
    h : list of tf.Tensor
        Hidden layers
    """

    def __init__(self,
                 layers, weights=None, init_weights=None, activation=None,
                 dtype='float', scope='global', do_setup=False):
        """Initialize Neural Network wrapper."""
        add_dependency(tf, 'TensorFlow')

        if (len(layers) < 2):
            raise ValueError('At least two layers needed.')

        # determine state and action space
        state_space = RdSpace((layers[0],))
        action_space = RdSpace((layers[-1],))

        # store arguments convenient for copy operation
        self.args = [layers]
        self.kwargs = {
            'weights': weights,
            'init_weights': init_weights,
            'activation': activation,
            'dtype': dtype
        }

        self.state_space = state_space
        self.action_space = action_space

        self.dtype = dtype
        self.layers = layers
        self.scope = scope

        self.is_set_up = False

        if init_weights is None:
            self.init_weights = default_init_weights
        else:
            self.init_weights = init_weights

        # Activation function
        if activation is None:
            activation = (len(layers) - 2) * [tf.sigmoid]
        elif (isinstance(activation, list)
                and (len(activation) != len(layers) - 2)):
            raise ValueError('Activation list has wrong size.')
        else:
            activation = (len(layers) - 2) * [activation]

        self.activation = activation

        # Symbols
        self.X = tf.placeholder(dtype, shape=[None, layers[0]], name='X')
        self.a = tf.placeholder(dtype, shape=[None, layers[-1]], name='a')

        if do_setup:
            with tf.variable_scope(self.scope):
                self.setup()
        else:
            # Make sure all fields exist
            self.W_action = None
            self.W_var = None
            self.a_pred = None
            self.var = None
            self.h = None

        self.sess = None

    def setup(self):
        """Setup the network graph.

        The weights and graph will be initialized by this function. If do_setup
        is True, setup will automatically be called, when instantiating the
        class.
        """
        if self.is_set_up:
            raise MultipleCallsException('Network is already set up.')

        layers = self.layers
        weights = self.kwargs['weights']

        # Weights for the action estimation
        with tf.variable_scope('action_estimator'):
            if weights is None:
                w = []
                for i in range(len(layers) - 1):
                    w.append(self.init_weights((layers[i], layers[i + 1])))
            else:
                w = weights

            self.W_action = w

        # generate network
        self.a_pred = self._generate_network()

        # Weights for variance estimation
        with tf.variable_scope('variance_estimator'):
            self.W_var = []
            for i in range(1, len(layers) - 1):
                self.W_var.append(self.init_weights((layers[i], 1)))

        # generate variance network
        self.var = self._generate_variance()

        self.is_set_up = True

    def _generate_network(self):
        self.h = [self.X]
        for i, act in enumerate(self.activation):
            h_i = self.h[i]
            w_i = self.W_action[i]
            self.h.append(act(tf.matmul(h_i, w_i)))

        return tf.matmul(self.h[-1], self.W_action[-1])

    def _generate_variance(self):
        var = []
        if not self.W_var:
            return tf.constant(0, name='variance')
        for h_i, w_i in zip(self.W_var, self.h[1:]):
            var.append(tf.reduce_sum(tf.matmul(w_i, h_i)))
        return tf.abs(tf.reduce_sum(var, name='variance'))

    def copy(self, scope, do_setup=True):
        """Generate a copy of the network.

        The copy will instantiate the class with the same arguments, but
        replace `scope` and `do_setup` with the respective arguments passed
        to this function.

        Parameters
        ----------
        scope : String
            Indication the scope that should be used when initializing the
            network.
        do_setup : Boolean
            Default: True ; Indicating if the `setup` method, should be called
            when instantiating.
        """
        self.kwargs['scope'] = scope
        self.kwargs['do_setup'] = do_setup
        return NeuralNetwork(*self.args, **self.kwargs)

    def map(self, state):
        """Compute output in session.

        Make sure a default session is set when calling.
        """
        state = state.flatten()
        assert(self.state_space.contains(state))

        if self.sess is None:
            sess = tf.get_default_session()
        else:
            sess = self.sess
        mean, var = sess.run([self.a_pred, self.var], {self.X: [state]})

        action = np.array(normal(mean, var))
        action = action.reshape(self.action_space.shape)

        return action

    @property
    def parameters(self):
        """Return weights of the neural network.

        This returns a list of tf.Variables. Please note that these can not
        simply be updated by assignment. See the parameters.setter docstring
        for more information.
        The list of tf.Variables can be directly accessed through the
        attribute `W`.
        """
        if self.sess is None:
            return tf.get_default_session().run(self.W_action + self.W_var)
        else:
            return self.sess.run(self.W_action + self.W_var)

    @parameters.setter
    def parameters(self, update):
        """Setter function for parameters.

        Since the parameters are a list of `tf.Variable`, we need to feed them
        into an assign operator. Thus the argument, needs to be a list
        containing an element for each Variable in `W_action` and `W_var` in
        that order, i.e. `W_var` will be the last element.

        Parameters
        ----------
        update :
            List of parameters for each `tf.Varible`

        Notes
        -----
        Make sure there is a default session or `self.sess` is set.
        """
        if not isinstance(update, list):
            update = [update]

        variables = self.W_action + self.W_var
        assign_op = []

        for (var, val) in zip(variables, update):
            assign_op.append(var.assign(val))

        if self.sess is None:
            sess = tf.get_default_session()
        else:
            sess = self.sess

        sess.run(assign_op)

    @property
    def parameter_space(self):
        """Return parameter space."""
        pass



"""Define Measurements."""

from abc import ABCMeta, abstractmethod
from six import add_metaclass

from operator import itemgetter

__all__ = ('Measure', 'BestPerformance', 'SafetyMeasure')


@add_metaclass(ABCMeta)
class Measure(object):
    """Abstract Base class defining the interface for any measurement.

    The methods below are abstract and need to be implemented by any child.

    Methods
    -------
    __call__(runs)
        Abstract! Evaluate a list of runs.
    result()
        Abstract! Return the result of the evaluation.
    """

    @abstractmethod
    def __call__(self, runs):
        """Evaluate a list of runs.

        Parameters
        ----------
        runs : List of BenchRun instances
            May be any subset of BenchRun instances passed in a list
        """
        pass

    @property
    @abstractmethod
    def result(self):
        """Return the result of evaluation."""
        pass


class BestPerformance(Measure):
    """Find the best performance achieved within runs."""

    def __call__(self, runs):
        """Sort content of runs by performance.

        This class creates a tuple with a BenchRun and its respective best
        performance and then stores in a descending sorted list.
        The results are accessible through the result method.

        Parameters
        ----------
        runs : List of BenchRun instances
            May be any subset of BenchRun instances in a list.
        """
        # create a list of tuples with the max reward for each run
        runs_tup = []
        for run in runs:
            monitor = run.get_alg_monitor()
            max_reward = max(monitor.rewards)
            runs_tup.append((run, max_reward))

        # sort list
        self._result = sorted(runs_tup, key=itemgetter(1), reverse=True)

    @property
    def result(self):
        """Retrieve result."""
        if not hasattr(self, '_result'):
            self._result = None
        return self._result

    @property
    def best_result(self):
        """Retrieve the best run."""
        if self.result is not None:
            return self.result[0]
        return None


class SafetyMeasure(Measure):
    """Detect Safety violations.

    The measure evaluates to a list of 3-tuples, where the first
    element contains the instance that was evaluated, the second one
    the number of violations that occured and the third the sum of
    those violations, i.e. the sum of the difference between the
    effective reward and the threshold, for every violation.

    Attributes
    ----------
    threshold : float or integer
        Reward threshold to detect violations.
    """

    def __init__(self, threshold):
        """Initialize SafetyMeasure.

        Parameters
        ----------
        threshold : float or integer
            Reward threshold to detect violations.
        """
        self.threshold = threshold

    def __call__(self, runs):
        """Evaluate Safety violations.

        Parameters
        ----------
        runs : List of BenchRun instances
            May be any subset of BenchRun instances in a list.
        """
        self._result = []

        for run in runs:
            num_violations = 0
            sum_violations = 0
            for reward in run.get_alg_monitor().rewards:
                if reward < self.threshold:
                    num_violations += 1
                    sum_violations += self.threshold - reward
            self._result.append((run, num_violations, sum_violations))

    @property
    def result(self):
        """Retrieve result.

        If a run has been evaluated, this function will retrieve the
        result otherwise it will return ``None``.
        The function evaluates to a list of 3-tuples, where the first
        element contains the instance that was evaluated, the second one
        the number of violations that occured and the third the sum of
        those violations, i.e. the sum of the difference between the
        effective reward and the threshold, for every violation.

        Returns
        -------
        (run, count, amount) : 3 tuple
            run: Object that has been evaluated
            count: Number of rollouts where the safety threshold was violated.
            amount: Sum of difference between violations and threshold.
        """
        if not hasattr(self, '_result'):
            self._result = None
        return self._result
